import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class SurfaceControlMass(om.ExplicitComponent):
    '''
    Calculate the mass of the surface controls. The methodology is based on the
    FLOPS weight equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER, val=1.0)
        add_aviary_input(self, Mission.Design.GROSS_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Wing.CONTROL_SURFACE_AREA_RATIO, val=0.0)
        add_aviary_input(self, Aircraft.Wing.AREA, val=0.0)

        add_aviary_output(self, Aircraft.Wing.SURFACE_CONTROL_MASS, val=10)
        add_aviary_output(self, Aircraft.Wing.CONTROL_SURFACE_AREA, val=2)

        self.declare_partials(Aircraft.Wing.SURFACE_CONTROL_MASS, '*')
        self.declare_partials(Aircraft.Wing.CONTROL_SURFACE_AREA, [
                              Aircraft.Wing.CONTROL_SURFACE_AREA_RATIO, Aircraft.Wing.AREA])

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        scaler = inputs[Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        flap_ratio = inputs[Aircraft.Wing.CONTROL_SURFACE_AREA_RATIO]
        wing_area = inputs[Aircraft.Wing.AREA]

        surface_flap_area = flap_ratio*wing_area

        surface_ctrls_wt = \
            1.1*max_mach**.52 * surface_flap_area**.6 * gross_weight**.32 * scaler

        outputs[Aircraft.Wing.CONTROL_SURFACE_AREA] = surface_flap_area

        outputs[Aircraft.Wing.SURFACE_CONTROL_MASS] = surface_ctrls_wt / \
            GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        scaler = inputs[Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)
        gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
        flap_ratio = inputs[Aircraft.Wing.CONTROL_SURFACE_AREA_RATIO]
        wing_area = inputs[Aircraft.Wing.AREA]

        surface_flap_area = flap_ratio*wing_area

        max_mach_exp = max_mach**.52
        surface_area_exp = surface_flap_area**.6
        gross_weight_exp = gross_weight**.32

        J[
            Aircraft.Wing.SURFACE_CONTROL_MASS,
            Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER] = \
            1.1*max_mach_exp * surface_area_exp * gross_weight_exp / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.SURFACE_CONTROL_MASS, Mission.Design.GROSS_MASS] = \
            1.1*max_mach_exp * surface_area_exp * .32 * \
            gross_weight**(.32-1) * scaler

        J[Aircraft.Wing.SURFACE_CONTROL_MASS, Aircraft.Wing.CONTROL_SURFACE_AREA_RATIO] = \
            1.1*max_mach_exp * .6 * \
            surface_flap_area**(.6-1) * gross_weight_exp * wing_area * \
            scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.SURFACE_CONTROL_MASS, Aircraft.Wing.AREA] = \
            1.1*max_mach_exp * .6 * \
            surface_flap_area**(.6-1) * gross_weight_exp * \
            flap_ratio * scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.CONTROL_SURFACE_AREA, Aircraft.Wing.CONTROL_SURFACE_AREA_RATIO] = \
            wing_area
        J[Aircraft.Wing.CONTROL_SURFACE_AREA, Aircraft.Wing.AREA] = flap_ratio


class AltSurfaceControlMass(om.ExplicitComponent):
    '''
    Calculate the mass of the surface controls using the alternate method.
    The methodology is based on the FLOPS weight equations, modified to
    output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER, val=1.0)
        add_aviary_input(self, Aircraft.Wing.AREA, val=0.0)
        add_aviary_input(self, Aircraft.Wing.CONTROL_SURFACE_AREA_RATIO, val=0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.WETTED_AREA, val=0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.THICKNESS_TO_CHORD, val=0.0)
        add_aviary_input(self, Aircraft.VerticalTail.AREA, val=0.0)

        add_aviary_output(self, Aircraft.Wing.CONTROL_SURFACE_AREA, val=2)
        add_aviary_output(self, Aircraft.Wing.SURFACE_CONTROL_MASS, val=0.0)

        self.declare_partials('*', '*')

    def compute(self, inputs, outputs):
        scaler = inputs[Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER]
        wing_area = inputs[Aircraft.Wing.AREA]
        flap_ratio = inputs[Aircraft.Wing.CONTROL_SURFACE_AREA_RATIO]
        htail_area = inputs[Aircraft.HorizontalTail.WETTED_AREA]
        htail_TCR = inputs[Aircraft.HorizontalTail.THICKNESS_TO_CHORD]
        vtail_area = inputs[Aircraft.VerticalTail.AREA]

        surface_ctrls_wt = (480 + .99*wing_area + 2.5*htail_area /
                            (2 + .387*htail_TCR) + 1.6*vtail_area) * scaler

        outputs[Aircraft.Wing.SURFACE_CONTROL_MASS] = surface_ctrls_wt / \
            GRAV_ENGLISH_LBM

        outputs[Aircraft.Wing.CONTROL_SURFACE_AREA] = flap_ratio * wing_area

    def compute_partials(self, inputs, J):
        scaler = inputs[Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER]
        wing_area = inputs[Aircraft.Wing.AREA]
        flap_ratio = inputs[Aircraft.Wing.CONTROL_SURFACE_AREA_RATIO]
        htail_area = inputs[Aircraft.HorizontalTail.WETTED_AREA]
        htail_TCR = inputs[Aircraft.HorizontalTail.THICKNESS_TO_CHORD]
        vtail_area = inputs[Aircraft.VerticalTail.AREA]

        J[
            Aircraft.Wing.SURFACE_CONTROL_MASS,
            Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER] = (
            480 + .99 * wing_area + 2.5 * htail_area / (2 + .387 * htail_TCR)
            + 1.6 * vtail_area) / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.SURFACE_CONTROL_MASS, Aircraft.Wing.AREA] = \
            .99 * scaler / GRAV_ENGLISH_LBM
        J[Aircraft.Wing.SURFACE_CONTROL_MASS, Aircraft.HorizontalTail.WETTED_AREA] = \
            2.5/(2 + .387*htail_TCR) * scaler / GRAV_ENGLISH_LBM

        J[
            Aircraft.Wing.SURFACE_CONTROL_MASS,
            Aircraft.HorizontalTail.THICKNESS_TO_CHORD] = \
            -2.5*htail_area/(2 + .387*htail_TCR)**2 * .387 * \
            scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.SURFACE_CONTROL_MASS, Aircraft.VerticalTail.AREA] = \
            1.6 * scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Wing.CONTROL_SURFACE_AREA, Aircraft.Wing.CONTROL_SURFACE_AREA_RATIO] = \
            wing_area

        J[Aircraft.Wing.CONTROL_SURFACE_AREA, Aircraft.Wing.AREA] = flap_ratio



import unittest
import os

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.mission.gasp_based.ode.descent_eom import DescentRates
from aviary.variable_info.variables import Dynamic


class DescentTestCase(unittest.TestCase):
    """
    These tests compare the output of the climb EOM to the output from GASP. There are some discrepancies.
    These discrepancies were considered to be small enough, given that the difference in calculation methods between the two codes is significant.
    """

    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group", DescentRates(num_nodes=2), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Dynamic.Mission.VELOCITY, np.array([459, 459]), units="kn")
        self.prob.model.set_input_defaults(
            Dynamic.Mission.THRUST_TOTAL, np.array([452, 452]), units="lbf")
        self.prob.model.set_input_defaults(
            Dynamic.Mission.DRAG, np.array([7966.927, 7966.927]), units="lbf"
        )  # estimated from GASP values
        self.prob.model.set_input_defaults(
            Dynamic.Mission.MASS, np.array([147661, 147661]), units="lbm"
        )
        self.prob.model.set_input_defaults("alpha", np.array([3.2, 3.2]), units="deg")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        tol = 1e-6
        self.prob.run_model()

        assert_near_equal(
            self.prob[Dynamic.Mission.ALTITUDE_RATE], np.array(
                [-39.41011217, -39.41011217]), tol
        )  # note: values from GASP are: np.array([-39.75, -39.75])
        assert_near_equal(
            self.prob[Dynamic.Mission.DISTANCE_RATE], np.array(
                [773.70165638, 773.70165638]), tol
            # note: these values are finite differenced and lose accuracy. Fd values are:np.array([964.4634921, 964.4634921])
        )
        assert_near_equal(
            self.prob["required_lift"],
            np.array([147444.58096139, 147444.58096139]),
            tol,
        )  # note: values from GASP are: np.array([146288.8, 146288.8]) (estimated based on GASP values)
        assert_near_equal(
            self.prob[Dynamic.Mission.FLIGHT_PATH_ANGLE], np.array(
                [-0.05089311, -0.05089311]), tol
        )  # note: values from GASP are: np.array([-.0513127, -.0513127])

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.constants import GRAV_ENGLISH_LBM, RHO_SEA_LEVEL_METRIC
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class LandingCalc(om.ExplicitComponent):
    def setup(self):

        add_aviary_input(self, Mission.Landing.TOUCHDOWN_MASS, val=150_000)

        self.add_input("rho", val=1.225, units="kg/m**3", desc="atmospheric density")

        add_aviary_input(self, Aircraft.Wing.AREA, val=700)

        add_aviary_input(self, Mission.Landing.LIFT_COEFFICIENT_MAX, val=3)

        add_aviary_output(self, Mission.Landing.GROUND_DISTANCE, val=0)

        add_aviary_output(self, Mission.Landing.INITIAL_VELOCITY, val=0)

        self.declare_partials(Mission.Landing.INITIAL_VELOCITY, [
                              Mission.Landing.LIFT_COEFFICIENT_MAX, Aircraft.Wing.AREA, Mission.Landing.TOUCHDOWN_MASS])
        self.declare_partials(Mission.Landing.GROUND_DISTANCE, "*")

    def compute(self, inputs, outputs):

        rho_SL = RHO_SEA_LEVEL_METRIC
        landing_weight = inputs[Mission.Landing.TOUCHDOWN_MASS] * \
            GRAV_ENGLISH_LBM
        rho = inputs["rho"]
        planform_area = inputs[Aircraft.Wing.AREA]
        Cl_ldg_max = inputs[Mission.Landing.LIFT_COEFFICIENT_MAX]

        rho_ratio = rho / rho_SL

        # TODO: This equation from FLOPS estimates the landing field length, not the actual ground
        # distance covered during landing, which should be less.

        Cl_app = Cl_ldg_max / 1.3 ** 2
        V_app = 17.18644 * (landing_weight / (planform_area * Cl_app)) ** 0.5
        landing_distance = 2500 + 105 * landing_weight / (
            planform_area * rho_ratio * Cl_app * 1.69
        )

        outputs[Mission.Landing.GROUND_DISTANCE] = landing_distance
        outputs[Mission.Landing.INITIAL_VELOCITY] = V_app

    def compute_partials(self, inputs, J):

        rho_SL = RHO_SEA_LEVEL_METRIC
        landing_weight = inputs[Mission.Landing.TOUCHDOWN_MASS] * \
            GRAV_ENGLISH_LBM
        rho = inputs["rho"]
        planform_area = inputs[Aircraft.Wing.AREA]
        Cl_ldg_max = inputs[Mission.Landing.LIFT_COEFFICIENT_MAX]

        rho_ratio = rho / rho_SL

        Cl_app = Cl_ldg_max / 1.3 ** 2

        J[Mission.Landing.INITIAL_VELOCITY, Mission.Landing.LIFT_COEFFICIENT_MAX] = (
            17.18644
            * 0.5
            * (landing_weight / (planform_area * Cl_app)) ** (-0.5)
            * (-landing_weight)
            / (planform_area * Cl_app ** 2)
            / 1.3 ** 2
        )
        J[Mission.Landing.INITIAL_VELOCITY, Aircraft.Wing.AREA] = (
            17.18644
            * 0.5
            * (landing_weight / (planform_area * Cl_app)) ** (-0.5)
            * (-landing_weight)
            / (planform_area ** 2 * Cl_app)
        )
        J[Mission.Landing.INITIAL_VELOCITY, Mission.Landing.TOUCHDOWN_MASS] = (
            17.18644
            * 0.5
            * (landing_weight / (planform_area * Cl_app)) ** (-0.5)
            * GRAV_ENGLISH_LBM / (planform_area * Cl_app)
        )

        J[Mission.Landing.GROUND_DISTANCE, Mission.Landing.TOUCHDOWN_MASS] = \
            105 * GRAV_ENGLISH_LBM / (
            planform_area * rho_ratio * Cl_app * 1.69
        )
        J[Mission.Landing.GROUND_DISTANCE, Aircraft.Wing.AREA] = (
            -105 * landing_weight / (planform_area ** 2 * rho_ratio * Cl_app * 1.69)
        )
        J[Mission.Landing.GROUND_DISTANCE, Mission.Landing.LIFT_COEFFICIENT_MAX] = (
            -105
            * landing_weight
            / (planform_area * rho_ratio * Cl_app ** 2 * 1.69)
            / 1.3 ** 2
        )
        J[Mission.Landing.GROUND_DISTANCE, "rho"] = (
            -105
            * landing_weight
            / (planform_area * rho_ratio ** 2 * Cl_app * 1.69)
            / rho_SL
        )


class LandingGroup(om.Group):
    def setup(self):

        self.add_subsystem(
            "USatm",
            USatm1976Comp(num_nodes=1),
            promotes_inputs=[("h", Mission.Landing.INITIAL_ALTITUDE)],
            promotes_outputs=["rho"],
        )

        self.add_subsystem(
            "calcs",
            LandingCalc(),
            promotes_inputs=[Mission.Landing.TOUCHDOWN_MASS, "rho",
                             Aircraft.Wing.AREA, Mission.Landing.LIFT_COEFFICIENT_MAX],
            promotes_outputs=[
                Mission.Landing.GROUND_DISTANCE,
                Mission.Landing.INITIAL_VELOCITY,
            ],
        )



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import (
    distributed_engine_count_factor, distributed_thrust_factor)
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft

# TODO: update other components to remove thse variables
_wing_ref_area = 'aircraft:wing:dimensions:area'
_engine_count_factor = 'aircraft:propulsion:control:engine_count_factor'
_max_scaled_thrust = 'aircraft:propulsion:dimensions:max_scaled_thrust'
_density_ratio = 'aircraft:fuel:weights:density_ratio'
_total_fuel_vol = 'TBD:total_fuel_vol:'


class TransportUnusableFuelMass(om.ExplicitComponent):
    '''
    Calculates the mass of unusable fuel using the default method.
    The methodology is based on the FLOPS weight equations, modified to
    output mass instead of weight.

    Assumptions
    -----------.
    All engines use fuel, single engine model on aircraft (both engine level thrust
    (scaled by num_engines) and propulsion-level count factor is being used)
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(
            self, Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.Fuel.DENSITY_RATIO, 1.0)

        add_aviary_input(self, Aircraft.Fuel.TOTAL_CAPACITY, 0.0)

        add_aviary_input(self, Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST, val=0.0)

        add_aviary_input(self, Aircraft.Wing.AREA, val=0.0)

        add_aviary_output(self, Aircraft.Fuel.TOTAL_VOLUME, val=0.0)

        add_aviary_output(self, Aircraft.Fuel.UNUSABLE_FUEL_MASS, val=0.0)

    def setup_partials(self):

        self.declare_partials(
            Aircraft.Fuel.TOTAL_VOLUME,
            [Aircraft.Fuel.TOTAL_CAPACITY, Aircraft.Fuel.DENSITY_RATIO])

        self.declare_partials(
            Aircraft.Fuel.UNUSABLE_FUEL_MASS,
            [
                Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER,
                Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST, Aircraft.Wing.AREA,
                Aircraft.Fuel.TOTAL_CAPACITY, Aircraft.Fuel.DENSITY_RATIO])

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        tank_count = aviary_options.get_val(Aircraft.Fuel.NUM_TANKS)
        scaler = inputs[Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER]
        density_ratio = inputs[Aircraft.Fuel.DENSITY_RATIO]
        total_capacity = inputs[Aircraft.Fuel.TOTAL_CAPACITY]
        num_eng = aviary_options.get_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES)
        num_eng_fact = distributed_engine_count_factor(num_eng)
        max_sls_thrust = inputs[Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST]
        thrust_factor = distributed_thrust_factor(max_sls_thrust, num_eng)
        wing_area = inputs[Aircraft.Wing.AREA]

        # TODO: check if this is really a volume
        # outputs[Aircraft.Fuel.TOTAL_VOLUME] = total_capacity / density_ratio

        outputs[Aircraft.Fuel.UNUSABLE_FUEL_MASS] = (
            (11.5 * num_eng_fact * thrust_factor**0.2 + 0.07 * wing_area + 1.6 * tank_count *
             total_capacity**0.28) * density_ratio) * scaler / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        tank_count = aviary_options.get_val(Aircraft.Fuel.NUM_TANKS)
        scaler = inputs[Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER]
        density_ratio = inputs[Aircraft.Fuel.DENSITY_RATIO]
        total_capacity = inputs[Aircraft.Fuel.TOTAL_CAPACITY]
        num_eng = aviary_options.get_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES)
        num_eng_fact = distributed_engine_count_factor(num_eng)
        max_sls_thrust = inputs[Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST]
        thrust_factor = distributed_thrust_factor(max_sls_thrust, num_eng)
        wing_area = inputs[Aircraft.Wing.AREA]

        term1 = thrust_factor**0.2
        term2 = total_capacity**0.28

        # J[Aircraft.Fuel.TOTAL_VOLUME, Aircraft.Fuel.TOTAL_CAPACITY] = (
        #     1.0 / density_ratio)

        # J[Aircraft.Fuel.TOTAL_VOLUME, Aircraft.Fuel.DENSITY_RATIO] = (
        #     -total_capacity / (density_ratio * density_ratio))

        J[
            Aircraft.Fuel.UNUSABLE_FUEL_MASS,
            Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER] = (
            (11.5 * num_eng_fact * term1 + 0.07 * wing_area + 1.6 * tank_count * term2)
            * density_ratio / GRAV_ENGLISH_LBM)

        J[Aircraft.Fuel.UNUSABLE_FUEL_MASS, Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST] = (
            2.3 * thrust_factor**-0.8 * density_ratio * scaler / GRAV_ENGLISH_LBM)

        J[Aircraft.Fuel.UNUSABLE_FUEL_MASS, Aircraft.Wing.AREA] = (
            0.07 * density_ratio * scaler / GRAV_ENGLISH_LBM)

        J[Aircraft.Fuel.UNUSABLE_FUEL_MASS, Aircraft.Fuel.TOTAL_CAPACITY] = (
            0.448 * tank_count * total_capacity**-0.72 * density_ratio * scaler / GRAV_ENGLISH_LBM)

        J[Aircraft.Fuel.UNUSABLE_FUEL_MASS, Aircraft.Fuel.DENSITY_RATIO] = (
            11.5 * num_eng_fact * term1 + 0.07 * wing_area + 1.6 * tank_count * term2) * scaler / GRAV_ENGLISH_LBM


class AltUnusableFuelMass(om.ExplicitComponent):
    '''
    Calculates the mass of unusable fuel using the alternate method.
    The methodology is based on the FLOPS weight equations, modified
    to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.Fuel.TOTAL_CAPACITY, 0.0)

        add_aviary_output(self, Aircraft.Fuel.UNUSABLE_FUEL_MASS, 0.0)

    def setup_partials(self):
        self.declare_partials(of='*', wrt='*')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        total_fuel_capacity = inputs[Aircraft.Fuel.TOTAL_CAPACITY]
        scaler = inputs[Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER]

        outputs[Aircraft.Fuel.UNUSABLE_FUEL_MASS] = (
            0.0084 * total_fuel_capacity * scaler
        )

    def compute_partials(self, inputs, J):

        total_fuel_capacity = inputs[Aircraft.Fuel.TOTAL_CAPACITY]
        scaler = inputs[Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER]

        J[
            Aircraft.Fuel.UNUSABLE_FUEL_MASS,
            Aircraft.Fuel.UNUSABLE_FUEL_MASS_SCALER] = (
            0.0084 * total_fuel_capacity
        )

        J[Aircraft.Fuel.UNUSABLE_FUEL_MASS, Aircraft.Fuel.TOTAL_CAPACITY] = (
            0.0084 * scaler
        )



import numpy as np
import openmdao.api as om

from aviary.subsystems.mass.flops_based.engine_pod import EnginePodMass
from aviary.subsystems.mass.flops_based.wing_common import (
    WingBendingMass, WingMiscMass, WingShearControlMass, WingTotalMass)
from aviary.subsystems.mass.flops_based.wing_detailed import DetailedWingBendingFact
from aviary.subsystems.mass.flops_based.wing_simple import SimpleWingBendingFact
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft


class WingMassGroup(om.Group):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        aviary_options: AviaryValues = self.options['aviary_options']

        self.add_subsystem('engine_pod_mass',
                           EnginePodMass(aviary_options=aviary_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])

        if Aircraft.Wing.INPUT_STATION_DIST in aviary_options:
            self.add_subsystem('wing_bending_factor',
                               DetailedWingBendingFact(aviary_options=aviary_options),
                               promotes_inputs=['*'], promotes_outputs=['*'])

        else:
            self.add_subsystem('wing_bending_factor',
                               SimpleWingBendingFact(aviary_options=aviary_options),
                               promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem('wing_misc',
                           WingMiscMass(aviary_options=aviary_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem('wing_shear_control',
                           WingShearControlMass(aviary_options=aviary_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem('wing_bending',
                           WingBendingMass(aviary_options=aviary_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem('wing_total',
                           WingTotalMass(aviary_options=aviary_options),
                           promotes_inputs=['*'], promotes_outputs=['*'])



import numpy as np
import openmdao.api as om

from aviary.variable_info.variables import Dynamic


class GammaComp(om.ExplicitComponent):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def initialize(self):
        self.options.declare("num_nodes", types=int)

    def setup(self):
        nn = self.options["num_nodes"]

        self.add_input("dh_dr", shape=nn, units="m/distance_units",
                       desc="change in altitude wrt range")

        self.add_input("d2h_dr2", shape=nn, units="m/distance_units**2",
                       desc="second derivative of altitude wrt range")

        self.add_output(Dynamic.Mission.FLIGHT_PATH_ANGLE, shape=nn, units="rad",
                        desc="flight path angle")

        self.add_output("dgam_dr", shape=nn, units="rad/distance_units",
                        desc="change in flight path angle per unit range traversed")

    def setup_partials(self):
        nn = self.options["num_nodes"]
        ar = np.arange(nn, dtype=int)

        self.declare_partials(of=Dynamic.Mission.FLIGHT_PATH_ANGLE,
                              wrt="dh_dr", rows=ar, cols=ar)
        self.declare_partials(of="dgam_dr", wrt=["dh_dr", "d2h_dr2"], rows=ar, cols=ar)

    def compute(self, inputs, outputs):
        dh_dr = inputs["dh_dr"]
        d2h_dr2 = inputs["d2h_dr2"]

        outputs[Dynamic.Mission.FLIGHT_PATH_ANGLE] = np.arctan(dh_dr)
        outputs["dgam_dr"] = d2h_dr2 / (dh_dr**2 + 1)

    def compute_partials(self, inputs, partials):
        dh_dr = inputs["dh_dr"]
        d2h_dr2 = inputs["d2h_dr2"]

        partials[Dynamic.Mission.FLIGHT_PATH_ANGLE, "dh_dr"] = 1. / (dh_dr**2 + 1)
        partials["dgam_dr", "dh_dr"] = -d2h_dr2 * dh_dr**2 / (dh_dr**2 + 1)**2
        partials["dgam_dr", "d2h_dr2"] = 1. / (dh_dr**2 + 1)



import unittest

import dymos as dm
import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import require_pyoptsparse, use_tempdirs

from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.variable_info.enums import SpeedType
from aviary.variable_info.variables import Aircraft, Mission, Dynamic
from aviary.constants import RHO_SEA_LEVEL_ENGLISH

from aviary.mission.gasp_based.ode.unsteady_solved.unsteady_solved_ode import UnsteadySolvedODE
from aviary.mission.gasp_based.ode.groundroll_ode import GroundrollODE
from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.mission.gasp_based.ode.unsteady_solved.unsteady_solved_ode import \
    UnsteadySolvedODE
from aviary.mission.gasp_based.phases.v_rotate_comp import VRotateComp
from aviary.variable_info.options import get_option_defaults
from aviary.interface.default_phase_info.solved import default_mission_subsystems
from aviary.variable_info.enums import SpeedType
from aviary.variable_info.variables import Aircraft, Dynamic


def run_mission(optimizer):
    mach = 0.78
    target_range = 2000.  # NM

    takeoff_mass = 175400  # lb
    climb_mach = 0.8
    cruise_alt = 37500  # ft
    cruise_mach = mach
    # target_range = 2000  # 3675  # NM
    solve_segments = False

    range_guesses = np.array([  # in meters
        0.,  # 'groundroll',
        1000.,  # 'rotation',
        1500.,  # 'ascent_to_gear_retract',
        2000.,  # 'ascent_to_flap_retract',
        2500.,  # 'ascent',
        3500.,  # Dynamic.Mission.VELOCITY_RATE,
        4500.,  # 'climb_at_constant_EAS',
        32000.,  # 'climb_at_constant_EAS_to_mach',
        300.e3,  # 'climb_at_constant_mach',
        600.e3,  # 'cruise',
        1700. * target_range,  # 'descent'
        1700. * target_range + 200000.,
    ])
    mass_guesses = np.array([
        1.0,
        0.999,
        0.999,
        0.999,
        0.998,
        0.998,
        0.998,
        0.990,
        0.969,
        0.951,
        0.873,
        0.866,
    ]) * takeoff_mass
    alt_guesses = [
        0.0,
        0.0,
        0.0,
        50.0,
        400.0,
        500.0,
        500.0,
        10000.,
        32857.,
        cruise_alt,
        cruise_alt,
        10000.,
    ]
    TAS_guesses = np.array([
        0.0,
        150.,
        200.,
        200.,
        200.,
        225.,
        251.,
        290.,
        465. * mach / 0.8,
        458. * mach / 0.8,
        483. * mach / 0.8,
        250.,
    ])

    mean_TAS = (TAS_guesses[1:] + TAS_guesses[:-1]) / 2. / 1.94
    range_durations = range_guesses[1:] - range_guesses[:-1]
    time_guesses = np.hstack((0., np.cumsum(range_durations / mean_TAS)))

    throttle_max = 1.0
    throttle_climb = 0.956
    throttle_cruise = 0.930
    throttle_idle = 0.0

    phase_info = {
        'groundroll': {
            'num_segments': 3,
            'throttle_setting': throttle_max,
            'input_speed_type': SpeedType.TAS,
            'ground_roll': True,
            'clean': False,
            'initial_ref': 100.,
            'initial_bounds': (0., 500.),
            'duration_ref': 100.,
            'duration_bounds': (50., 1000.),
            'control_order': 1,
            'opt': True,
        },
        'rotation': {
            'num_segments': 2,
            'throttle_setting': throttle_max,
            'input_speed_type': SpeedType.TAS,
            'ground_roll': True,
            'clean': False,
            'initial_ref': 1.e3,
            'initial_bounds': (50., 5000.),
            'duration_ref': 1.e3,
            'duration_bounds': (50., 2000.),
            'control_order': 1,
            'opt': True,
        },
        'ascent_to_gear_retract': {
            'num_segments': 2,
            'throttle_setting': throttle_max,
            'input_speed_type': SpeedType.TAS,
            'ground_roll': False,
            'clean': False,
            'initial_ref': 1.e3,
            'initial_bounds': (10., 2.e3),
            'duration_ref': 1.e3,
            'duration_bounds': (500., 1.e4),
            'control_order': 1,
            'opt': False,
        },
        'ascent_to_flap_retract': {
            'num_segments': 2,
            'throttle_setting': throttle_max,
            'input_speed_type': SpeedType.TAS,
            'ground_roll': False,
            'clean': False,
            'initial_ref': 1.e3,
            'initial_bounds': (10., 2.e5),
            'duration_ref': 1.e3,
            'duration_bounds': (500., 1.e4),
            'control_order': 1,
            'opt': False,
        },
        'ascent': {
            'num_segments': 2,
            'throttle_setting': throttle_max,
            'input_speed_type': SpeedType.TAS,
            'ground_roll': False,
            'clean': True,
            'initial_ref': 1.e3,
            'initial_bounds': (10., 2.e5),
            'duration_ref': 1.e3,
            'duration_bounds': (500., 1.e4),
            'control_order': 1,
            'opt': False,
        },
        Dynamic.Mission.VELOCITY_RATE: {
            'num_segments': 2,
            'throttle_setting': throttle_max,
            'input_speed_type': SpeedType.TAS,
            'ground_roll': False,
            'clean': True,
            'initial_ref': 1.e3,
            'initial_bounds': (10., 2.e5),
            'duration_ref': 1.e3,
            'duration_bounds': (500., 1.e4),
            'control_order': 1,
            'opt': False,
        },
        'climb_at_constant_EAS': {
            'num_segments': 2,
            'throttle_setting': throttle_climb,
            'input_speed_type': SpeedType.EAS,
            'ground_roll': False,
            'clean': True,
            'initial_ref': 1.e5,
            'initial_bounds': (100., 1.e7),
            'duration_ref': 1.e5,
            'duration_bounds': (1.e4, 5.e4),
            'control_order': 1,
            'opt': False,
            'fixed_EAS': 250.,
        },
        'climb_at_constant_EAS_to_mach': {
            'num_segments': 2,
            'throttle_setting': throttle_climb,
            'input_speed_type': SpeedType.EAS,
            'ground_roll': False,
            'clean': True,
            'initial_ref': 1.e5,
            'initial_bounds': (100., 1.e7),
            'duration_ref': 1.e5,
            'duration_bounds': (10.e3, 1.e6),
            'control_order': 1,
            'opt': True,
            'fixed_EAS': 270.,
        },
        'climb_at_constant_mach': {
            'num_segments': 2,
            'throttle_setting': throttle_climb,
            'input_speed_type': SpeedType.MACH,
            'ground_roll': False,
            'clean': True,
            'initial_ref': 1.e5,
            'initial_bounds': (100., 1.e7),
            'duration_ref': 1.e5,
            'duration_bounds': (10.e3, 1.e6),
            'control_order': 1,
            'opt': True,
        },
        'cruise': {
            'num_segments': 5,
            'throttle_setting': throttle_cruise,
            'input_speed_type': SpeedType.MACH,
            'ground_roll': False,
            'clean': True,
            'initial_ref': 1.e6,
            'initial_bounds': (1.e4, 2.e7),
            'duration_ref': 5.e6,
            'duration_bounds': (0.1e7, 3.e7),
            'control_order': 1,
            'opt': False,
        },
        'descent': {
            'num_segments': 3,
            'throttle_setting': throttle_idle,
            'input_speed_type': SpeedType.TAS,
            'ground_roll': False,
            'clean': True,
            'initial_ref': 1.e6,
            'initial_bounds': (100., 1.e7),
            'duration_ref': 1.e5,
            'duration_bounds': (1.5e5, 2.e5),
            'control_order': 1,
            'opt': True,
        },
    }
    phases = list(phase_info.keys())

    p = om.Problem()

    traj = p.model.add_subsystem("traj", dm.Trajectory())

    for idx, phase_name in enumerate(phases):
        throttle_setting = phase_info[phase_name]['throttle_setting']
        input_speed_type = phase_info[phase_name]['input_speed_type']
        num_segments = phase_info[phase_name]['num_segments']
        ground_roll = phase_info[phase_name]['ground_roll']
        clean = phase_info[phase_name]['clean']
        initial_bounds = phase_info[phase_name]['initial_bounds']
        duration_bounds = phase_info[phase_name]['duration_bounds']
        initial_ref = phase_info[phase_name]['initial_ref']
        duration_ref = phase_info[phase_name]['duration_ref']
        control_order = phase_info[phase_name]['control_order']
        opt = phase_info[phase_name]['opt']

        if phase_name == "groundroll":
            groundroll_trans = dm.Radau(
                num_segments=num_segments, order=3, compressed=True, solve_segments="forward",
            )

            ode_args = dict(
                aviary_options=get_option_defaults(),
                core_subsystems=default_mission_subsystems
            )

            phase = dm.Phase(
                ode_class=GroundrollODE,
                ode_init_kwargs=ode_args,
                transcription=groundroll_trans,
            )

            phase.set_time_options(fix_initial=True, fix_duration=False,
                                   units="kn", name=Dynamic.Mission.VELOCITY,
                                   duration_bounds=duration_bounds, duration_ref=duration_ref, initial_ref=initial_ref)

            phase.set_state_options("time", rate_source="dt_dv", units="s",
                                    fix_initial=True, fix_final=False, ref=1., defect_ref=1.)

            phase.set_state_options("mass", rate_source="dmass_dv",
                                    fix_initial=True, fix_final=False, lower=1, upper=195_000, ref=takeoff_mass, defect_ref=takeoff_mass)

            phase.set_state_options(Dynamic.Mission.DISTANCE, rate_source="over_a",
                                    fix_initial=True, fix_final=False, lower=0, upper=2000., ref=1.e2, defect_ref=1.e2)

            phase.add_parameter("t_init_gear", units="s",
                                static_target=True, opt=False, val=32.3)
            phase.add_parameter("t_init_flaps", units="s",
                                static_target=True, opt=False, val=44.0)
            phase.add_parameter("wing_area", units="ft**2",
                                static_target=True, opt=False, val=1370)

        else:

            trans = dm.Radau(num_segments=num_segments, order=3,
                             compressed=True, solve_segments=solve_segments)

            ode_args = dict(
                input_speed_type=input_speed_type,
                clean=clean,
                ground_roll=ground_roll,
                aviary_options=get_option_defaults(),
                core_subsystems=default_mission_subsystems
            )

            phase = dm.Phase(
                ode_class=UnsteadySolvedODE,
                ode_init_kwargs=ode_args,
                transcription=trans,
            )

            phase.add_parameter(
                Dynamic.Mission.THROTTLE,
                opt=False,
                units="unitless",
                val=throttle_setting,
                static_target=False)

            phase.set_time_options(fix_initial=False, fix_duration=False,
                                   units="distance_units", name=Dynamic.Mission.DISTANCE,
                                   duration_bounds=duration_bounds, duration_ref=duration_ref,
                                   initial_bounds=initial_bounds, initial_ref=initial_ref)

            if phase_name == "cruise" or phase_name == "descent":
                time_ref = 1.e4
            else:
                time_ref = 100.

            phase.set_state_options("time", rate_source="dt_dr", targets=['t_curr'] if 'retract' in phase_name else [],
                                    fix_initial=False, fix_final=False, ref=time_ref, defect_ref=time_ref * 1.e2)

            phase.set_state_options("mass", rate_source="dmass_dr",
                                    fix_initial=False, fix_final=False, ref=170.e3, defect_ref=170.e5,
                                    val=170.e3, units='lbm', lower=10.e3)

            phase.add_parameter("wing_area", units="ft**2",
                                static_target=True, opt=False, val=1370)

            if Dynamic.Mission.VELOCITY_RATE in phase_name or 'ascent' in phase_name:
                phase.add_parameter(
                    "t_init_gear", units="s", static_target=True, opt=False, val=100)
                phase.add_parameter(
                    "t_init_flaps", units="s", static_target=True, opt=False, val=100)

            if 'rotation' in phase_name:
                phase.add_polynomial_control("TAS",
                                             order=control_order,
                                             units="kn", val=200.0,
                                             opt=opt, lower=1, upper=500, ref=250)

                phase.add_polynomial_control(Dynamic.Mission.ALTITUDE,
                                             order=control_order,
                                             fix_initial=False,
                                             rate_targets=['dh_dr'], rate2_targets=['d2h_dr2'],
                                             opt=False, upper=40.e3, ref=30.e3, lower=-1.)

                phase.add_polynomial_control("alpha",
                                             order=control_order,
                                             lower=-4, upper=15,
                                             units='deg', ref=10.,
                                             val=[0., 5.],
                                             opt=opt)
            else:
                if 'constant_EAS' in phase_name:
                    fixed_EAS = phase_info[phase_name]['fixed_EAS']
                    phase.add_parameter("EAS", units="kn", val=fixed_EAS)
                elif 'constant_mach' in phase_name:
                    phase.add_parameter(
                        Dynamic.Mission.MACH,
                        units="unitless",
                        val=climb_mach)
                elif 'cruise' in phase_name:
                    phase.add_parameter(
                        Dynamic.Mission.MACH, units="unitless", val=cruise_mach)
                else:
                    phase.add_polynomial_control("TAS",
                                                 order=control_order,
                                                 fix_initial=False,
                                                 units="kn", val=200.0,
                                                 opt=True, lower=1, upper=500, ref=250)

                phase.add_polynomial_control(Dynamic.Mission.ALTITUDE,
                                             order=control_order,
                                             units="ft", val=0.,
                                             fix_initial=False,
                                             rate_targets=['dh_dr'], rate2_targets=['d2h_dr2'],
                                             opt=opt, upper=40.e3, ref=30.e3, lower=-1.)

        if phase_name in phases[1:3]:
            phase.add_path_constraint(
                "fuselage_pitch", upper=15., units='deg', ref=15)
        if phase_name == "rotation":
            phase.add_boundary_constraint(
                "TAS", loc="final", upper=200., units="kn", ref=200.)
            phase.add_boundary_constraint(
                "normal_force", loc="final", equals=0., units="lbf", ref=10000.0)
        elif phase_name == "ascent_to_gear_retract":
            phase.add_path_constraint("load_factor", lower=0.0, upper=1.10)
        elif phase_name == "ascent_to_flap_retract":
            phase.add_path_constraint("load_factor", lower=0.0, upper=1.10)
        elif phase_name == "ascent":
            phase.add_path_constraint("EAS", upper=250., units="kn", ref=250.)
        elif phase_name == Dynamic.Mission.VELOCITY_RATE:
            phase.add_boundary_constraint(
                "EAS", loc="final", equals=250., units="kn", ref=250.)
        elif phase_name == "climb_at_constant_EAS":
            pass
        elif phase_name == "climb_at_constant_EAS_to_mach":
            phase.add_boundary_constraint(
                Dynamic.Mission.MACH, loc="final", equals=climb_mach, units="unitless")
        elif phase_name == "climb_at_constant_mach":
            pass
        elif phase_name == "descent":
            phase.add_boundary_constraint(
                Dynamic.Mission.DISTANCE,
                loc="final",
                equals=target_range,
                units="NM",
                ref=1.e3)
            phase.add_boundary_constraint(
                Dynamic.Mission.ALTITUDE,
                loc="final",
                equals=10.e3,
                units="ft",
                ref=10e3)
            phase.add_boundary_constraint(
                "TAS", loc="final", equals=250., units="kn", ref=250.)

        phase.add_timeseries_output(Dynamic.Mission.THRUST_TOTAL, units="lbf")
        phase.add_timeseries_output("thrust_req", units="lbf")
        phase.add_timeseries_output("normal_force")
        phase.add_timeseries_output(Dynamic.Mission.MACH)
        phase.add_timeseries_output("EAS", units="kn")
        phase.add_timeseries_output("TAS", units="kn")
        phase.add_timeseries_output(Dynamic.Mission.LIFT)
        phase.add_timeseries_output("CL")
        phase.add_timeseries_output("CD")
        phase.add_timeseries_output("time")
        phase.add_timeseries_output("mass")
        phase.add_timeseries_output(Dynamic.Mission.ALTITUDE)
        phase.add_timeseries_output("gear_factor")
        phase.add_timeseries_output("flap_factor")
        phase.add_timeseries_output("alpha")
        phase.add_timeseries_output(
            "fuselage_pitch", output_name="theta", units="deg")

        if 'rotation' in phase_name:
            phase.add_parameter("t_init_gear", units="s",
                                static_target=True, opt=False, val=100)
            phase.add_parameter("t_init_flaps", units="s",
                                static_target=True, opt=False, val=100)

        traj.add_phase(phase_name, phase)

    traj.add_linkage_constraint(phase_a='ascent_to_gear_retract',
                                phase_b='ascent_to_flap_retract',
                                var_a='time',
                                var_b='t_init_gear',
                                loc_a='final',
                                loc_b='initial',
                                connected=True)

    traj.add_linkage_constraint(phase_a='ascent_to_gear_retract',
                                phase_b='ascent',
                                var_a='time',
                                var_b='t_init_gear',
                                loc_a='final',
                                loc_b='initial',
                                connected=True)

    traj.add_linkage_constraint(phase_a='ascent_to_gear_retract',
                                phase_b=Dynamic.Mission.VELOCITY_RATE,
                                var_a='time',
                                var_b='t_init_gear',
                                loc_a='final',
                                loc_b='initial',
                                connected=True)

    traj.add_linkage_constraint(phase_a='ascent_to_flap_retract',
                                phase_b='ascent',
                                var_a='time',
                                var_b='t_init_flaps',
                                loc_a='final',
                                loc_b='initial',
                                connected=True)

    traj.add_linkage_constraint(phase_a='ascent_to_flap_retract',
                                phase_b=Dynamic.Mission.VELOCITY_RATE,
                                var_a='time',
                                var_b='t_init_flaps',
                                loc_a='final',
                                loc_b='initial',
                                connected=True)

    traj.link_phases(phases[6:], vars=[Dynamic.Mission.ALTITUDE], ref=10.e3)
    traj.link_phases(phases, vars=['time'], ref=100.)
    traj.link_phases(phases, vars=['mass'], ref=10.e3)
    traj.link_phases(phases, vars=[Dynamic.Mission.DISTANCE], units='m', ref=10.e3)
    traj.link_phases(phases[1:7], vars=['TAS'], units='kn', ref=200.)
    traj.add_linkage_constraint(phase_a='groundroll',
                                phase_b='rotation',
                                var_a=Dynamic.Mission.VELOCITY,
                                var_b='TAS',
                                loc_a='final',
                                loc_b='initial',
                                connected=False)

    p.model.add_subsystem("vrot_comp", VRotateComp())
    p.model.connect('traj.groundroll.states:mass',
                    'vrot_comp.mass', src_indices=om.slicer[0, ...])

    vrot_eq_comp = p.model.add_subsystem("vrot_eq_comp", om.EQConstraintComp())
    vrot_eq_comp.add_eq_output("v_rotate_error", eq_units="kn",
                               lhs_name="v_rot_computed", rhs_name="groundroll_v_final", add_constraint=True)

    p.model.connect('vrot_comp.Vrot', 'vrot_eq_comp.v_rot_computed')
    p.model.connect('traj.groundroll.timeseries.velocity',
                    'vrot_eq_comp.groundroll_v_final', src_indices=om.slicer[-1, ...])

    traj.add_parameter('wing_area', units='ft**2', static_target=True, opt=False)

    p.driver = om.pyOptSparseDriver()
    p.driver.declare_coloring(show_sparsity=False, show_summary=False)
    p.driver.options["print_results"] = p.comm.rank == 0

    print_opt_iters = True

    p.driver.options["optimizer"] = optimizer

    if optimizer == "SNOPT":
        p.driver.opt_settings["Major optimality tolerance"] = 1e-6
        p.driver.opt_settings["Major feasibility tolerance"] = 1e-6
        # p.driver.opt_settings["Function precision"] = 1e-6
        # p.driver.opt_settings["Linesearch tolerance"] = 0.99
        # p.driver.opt_settings["Major step limit"] = 0.1
        if print_opt_iters:
            p.driver.opt_settings["iSumm"] = 6
        p.driver.opt_settings["Major iterations limit"] = 50
    elif optimizer == "IPOPT":
        p.driver.opt_settings["max_iter"] = 100
        p.driver.opt_settings["tol"] = 1e-6
        if print_opt_iters:
            p.driver.opt_settings["print_level"] = 5
        p.driver.opt_settings[
            "nlp_scaling_method"
        ] = "gradient-based"  # for faster convergence
        p.driver.opt_settings["alpha_for_y"] = "safer-min-dual-infeas"
        p.driver.opt_settings["mu_strategy"] = "monotone"
        p.driver.opt_settings["bound_mult_init_method"] = "mu-based"

    p.set_solver_print(level=-1)

    obj_comp = om.ExecComp(f"obj = -final_mass / {takeoff_mass} + final_time / 5.",
                           final_mass={"units": "lbm"},
                           final_time={"units": "h"})
    p.model.add_subsystem("obj_comp", obj_comp)

    final_phase_name = phases[-1]
    p.model.connect(f"traj.{final_phase_name}.timeseries.mass",
                    "obj_comp.final_mass", src_indices=[-1])
    p.model.connect(f"traj.{final_phase_name}.timeseries.time",
                    "obj_comp.final_time", src_indices=[-1])

    p.model.add_objective("obj_comp.obj", ref=1.0)

    p.setup()

    # SET INPUTS FOR POST TRAJECTORY ANALYSES
    # TODO: paramport
    params = ParamPort.param_data
    p.set_val('vrot_comp.' + Aircraft.Wing.AREA,
              params[Aircraft.Wing.AREA]["val"], units=params[Aircraft.Wing.AREA]["units"])
    p.set_val("vrot_comp.dV1", val=10, units="kn")
    p.set_val("vrot_comp.dVR", val=5, units="kn")
    p.set_val("vrot_comp.rho", val=RHO_SEA_LEVEL_ENGLISH, units="slug/ft**3")
    p.set_val("vrot_comp.CL_max", val=2.1886, units="unitless")

    p.set_val("traj.parameters:wing_area", val=1370, units="ft**2")

    for idx, (phase_name, phase) in enumerate(traj._phases.items()):
        if phase_name != "groundroll":
            distance_initial = range_guesses[idx]
            p.set_val(f"traj.{phase_name}.t_initial",
                      distance_initial, units='distance_units')
            p.set_val(f"traj.{phase_name}.t_duration",
                      range_guesses[idx+1] - distance_initial, units='distance_units')

            p.set_val(
                f"traj.{phase_name}.polynomial_controls:altitude",
                phase.interp(Dynamic.Mission.ALTITUDE, [
                             alt_guesses[idx], alt_guesses[idx + 1]]),
                units="ft",
            )

            if "constant_EAS" in phase_name:
                pass
            elif "constant_mach" in phase_name:
                pass
            elif "cruise" in phase_name:
                pass
            else:
                p.set_val(
                    f"traj.{phase_name}.polynomial_controls:TAS",
                    phase.interp("TAS", [
                                 TAS_guesses[idx], TAS_guesses[idx+1]]),
                    units="kn",
                )
        else:
            p.set_val(f"traj.{phase_name}.t_initial", 0., units='kn')
            p.set_val(f"traj.{phase_name}.t_duration", 100., units='kn')

        p.set_val(
            f"traj.{phase_name}.states:mass",
            phase.interp("mass", [mass_guesses[idx], mass_guesses[idx+1]]),
            units="lbm",
        )

        p.set_val(
            f"traj.{phase_name}.states:time",
            phase.interp("time", [time_guesses[idx], time_guesses[idx+1]]),
            units="s",
        )

    dm.run_problem(p, run_driver=True, simulate=False, make_plots=False,
                   solution_record_file=f'solved_{optimizer}.sql',
                   )

    ranges = []
    masses = []
    alts = []
    TASs = []
    for idx, (phase_name, phase) in enumerate(traj._phases.items()):
        if phase_name == "groundroll":

            ranges.extend(
                p.get_val(f"traj.{phase_name}.timeseries.distance", units="m")[0])
            ranges.extend(
                p.get_val(f"traj.{phase_name}.timeseries.distance", units="m")[-1])

            masses.extend(
                p.get_val(f"traj.{phase_name}.timeseries.mass", units="lbm")[0])
            masses.extend(
                p.get_val(f"traj.{phase_name}.timeseries.mass", units="lbm")[-1])

            alts.extend([0., 0.])

            TASs.extend(
                p.get_val(f"traj.{phase_name}.timeseries.velocity", units="kn")[0])
            TASs.extend(
                p.get_val(f"traj.{phase_name}.timeseries.velocity", units="kn")[-1])
        else:
            distance_initial = p.get_val(
                f"traj.{phase_name}.t_initial", units='distance_units')
            if idx > 1:
                ranges.extend(distance_initial)
            if idx == (len(traj._phases) - 1):
                ranges.extend(
                    p.get_val(f"traj.{phase_name}.t_duration", units='distance_units') + distance_initial)
            masses.extend(
                p.get_val(f"traj.{phase_name}.timeseries.mass", units="lbm")[-1])
            alts.extend(
                p.get_val(f"traj.{phase_name}.timeseries.altitude", units="ft")[-1])
            TASs.extend(
                p.get_val(f"traj.{phase_name}.timeseries.TAS", units="kn")[-1])

    return p


@use_tempdirs
class TestFullMission(unittest.TestCase):

    def assert_result(self, p):
        tf = p.get_val('traj.descent.timeseries.time', units='s')[-1, 0]
        vf = p.get_val('traj.descent.timeseries.TAS', units='kn')[-1, 0]
        wf = p.get_val('traj.descent.timeseries.mass', units='lbm')[-1, 0]

        print(f't_final: {tf:8.3f} s')
        print(f'v_final: {vf:8.3f} knots')
        print(f'w_final: {wf:8.3f} lbm')

        assert_near_equal(tf, 16234., tolerance=0.02)
        assert_near_equal(vf, 250., tolerance=0.02)
        assert_near_equal(wf, 150978., tolerance=0.02)

    @require_pyoptsparse(optimizer='IPOPT')
    def bench_test_rotation_result_ipopt(self):
        p = run_mission(optimizer='IPOPT')
        self.assert_result(p)

    @require_pyoptsparse(optimizer='SNOPT')
    def bench_test_rotation_result_snopt(self):
        p = run_mission(optimizer='SNOPT')
        self.assert_result(p)


if __name__ == '__main__':
    use_SNOPT = True

    z = TestFullMission()

    if use_SNOPT:
        z.bench_test_rotation_result_snopt()

    else:
        z.bench_test_rotation_result_ipopt()



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.variable_info.enums import AnalysisScheme
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic


class FlightConstraints(om.ExplicitComponent):

    """
    Compute the minimum TAS (defined as the stall speed multiplied by a safety factor of
    1.1).

    Also compute the fuselage pitch angle.

    Both equations come from the climb subroutine of GASP.
    """

    def initialize(self):
        self.options.declare("analysis_scheme", types=AnalysisScheme,
                             default=AnalysisScheme.COLLOCATION)
        self.options.declare("num_nodes", types=int)
        self.options.declare("alt_trigger_units", default="ft")
        self.options.declare("speed_trigger_units", default="kn")

    def setup(self):
        analysis_scheme = self.options["analysis_scheme"]
        nn = self.options["num_nodes"]
        arange = np.arange(nn)

        if analysis_scheme is AnalysisScheme.SHOOTING:
            self.add_input(
                "alt_trigger", units=self.options["alt_trigger_units"], val=10.e3)
            self.add_input("speed_trigger",
                           units=self.options["speed_trigger_units"], val=100)
        self.add_input(
            Dynamic.Mission.MASS,
            val=np.ones(nn),
            units="lbm",
            desc="mass of aircraft",
        )

        add_aviary_input(self, Aircraft.Wing.AREA, val=2)

        self.add_input(
            "rho",
            val=np.ones(nn),
            units="slug/ft**3",
            desc="density of air",
        )
        self.add_input(
            "CL_max",
            val=np.ones(nn),
            units="unitless",
            desc="maximum lift coefficient",
        )
        self.add_input(
            Dynamic.Mission.FLIGHT_PATH_ANGLE,
            val=np.ones(nn),
            units="rad",
            desc="flight path angle",
        )

        add_aviary_input(self, Aircraft.Wing.INCIDENCE, val=1.0, units="rad")

        self.add_input(
            "alpha",
            val=np.ones(nn),
            units="rad",
            desc="angle of attack",
        )
        self.add_input(
            "TAS",
            val=np.ones(nn),
            units="ft/s",
            desc="true airspeed",
        )

        self.add_output(
            "theta",
            val=np.ones(nn),
            units="rad",
            desc="pitch angle of fuselage",
        )
        self.add_output(
            "TAS_violation",
            val=np.ones(nn),
            units="ft/s",
            desc="value to show if minimum TAS constraint is being violated. Negative or"
            " zero if constraint is satisfied.",
        )
        self.add_output("TAS_min", val=np.zeros(nn), units="ft/s")

        self.declare_partials(
            "theta", [Dynamic.Mission.FLIGHT_PATH_ANGLE, "alpha"], rows=arange, cols=arange)
        self.declare_partials(
            "theta",
            [
                Aircraft.Wing.INCIDENCE,
            ],
        )
        self.declare_partials(
            "TAS_violation",
            [Dynamic.Mission.MASS, "rho", "CL_max", "TAS"],
            rows=arange,
            cols=arange,
        )
        self.declare_partials(
            "TAS_violation",
            [
                Aircraft.Wing.AREA,
            ],
        )
        self.declare_partials(
            "TAS_min", [Dynamic.Mission.MASS, "rho", "CL_max"], rows=arange, cols=arange
        )
        self.declare_partials(
            "TAS_min",
            [
                Aircraft.Wing.AREA,
            ],
        )

    def compute(self, inputs, outputs):

        weight = inputs[Dynamic.Mission.MASS] * GRAV_ENGLISH_LBM
        wing_area = inputs[Aircraft.Wing.AREA]
        rho = inputs["rho"]
        CL_max = inputs["CL_max"]
        gamma = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]
        i_wing = inputs[Aircraft.Wing.INCIDENCE]
        alpha = inputs["alpha"]
        TAS = inputs["TAS"]

        V_stall = (2 * weight / (wing_area * rho * CL_max)) ** 0.5  # stall speed
        TAS_min = (
            1.1 * V_stall
        )  # minimum true airspeed accross each node, based on stall speed and safety margin
        outputs["TAS_min"] = TAS_min

        outputs["theta"] = gamma - i_wing + alpha
        outputs["TAS_violation"] = TAS_min - TAS

    def compute_partials(self, inputs, J):

        weight = inputs[Dynamic.Mission.MASS] * GRAV_ENGLISH_LBM
        wing_area = inputs[Aircraft.Wing.AREA]
        rho = inputs["rho"]
        CL_max = inputs["CL_max"]
        gamma = inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE]
        i_wing = inputs[Aircraft.Wing.INCIDENCE]
        alpha = inputs["alpha"]
        TAS = inputs["TAS"]

        J["theta", Dynamic.Mission.FLIGHT_PATH_ANGLE] = 1
        J["theta", "alpha"] = 1
        J["theta", Aircraft.Wing.INCIDENCE] = -1

        J["TAS_violation", Dynamic.Mission.MASS] = (
            1.1 * 0.5 * (2 / (wing_area * rho * CL_max)) ** 0.5 *
            weight ** (-0.5) * GRAV_ENGLISH_LBM
        )
        J["TAS_violation", "rho"] = (
            1.1 * (2 * weight / (wing_area * CL_max)) ** 0.5 * (-0.5) * rho ** (-1.5)
        )
        J["TAS_violation", "CL_max"] = (
            1.1 * (2 * weight / (wing_area * rho)) ** 0.5 * (-0.5) * CL_max ** (-1.5)
        )
        J["TAS_violation", "TAS"] = -1
        J["TAS_violation", Aircraft.Wing.AREA] = (
            1.1 * (2 * weight / (rho * CL_max)) ** 0.5 * (-0.5) * wing_area ** (-1.5)
        )

        J["TAS_min", Dynamic.Mission.MASS] = 1.1 * (
            0.5 * (2 / (wing_area * rho * CL_max)) ** 0.5 *
            weight ** (-0.5) * GRAV_ENGLISH_LBM
        )
        J["TAS_min", "rho"] = 1.1 * (
            (2 * weight / (wing_area * CL_max)) ** 0.5 * (-0.5) * rho ** (-1.5)
        )
        J["TAS_min", "CL_max"] = 1.1 * (
            (2 * weight / (wing_area * rho)) ** 0.5 * (-0.5) * CL_max ** (-1.5)
        )
        J["TAS_min", Aircraft.Wing.AREA] = 1.1 * (
            (2 * weight / (rho * CL_max)) ** 0.5 * (-0.5) * wing_area ** (-1.5)
        )


class ClimbAtTopOfClimb(om.ExplicitComponent):
    def setup(self):
        self.add_input("TAS", units="ft/s", val=-200)
        self.add_input(
            Dynamic.Mission.FLIGHT_PATH_ANGLE, units="rad", val=0.)
        self.add_output("ROC", units="ft/s")
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        outputs["ROC"] = inputs["TAS"] * \
            np.sin(inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE])

    def compute_partials(self, inputs, J):
        J["ROC", "TAS"] = np.sin(
            inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE])
        J["ROC", Dynamic.Mission.FLIGHT_PATH_ANGLE] = inputs["TAS"] * \
            np.cos(inputs[Dynamic.Mission.FLIGHT_PATH_ANGLE])



import numpy as np
import openmdao.api as om
from aviary.constants import GRAV_ENGLISH_LBM


class MassToWeight(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int, default=1)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(
            "mass",
            val=np.ones(nn),
            units="lbm",
            desc="mass of the aircraft",
        )

        self.add_output(
            "weight",
            val=np.ones(nn),
            units="lbf",
            desc="weight of the aircraft",
        )

    def setup_partials(self):
        nn = self.options['num_nodes']
        arange = np.arange(nn)
        self.declare_partials("weight",
                              "mass",
                              rows=arange,
                              cols=arange,
                              val=np.full(nn, GRAV_ENGLISH_LBM)
                              )

    def compute(self, inputs, outputs):
        outputs["weight"] = inputs["mass"] * GRAV_ENGLISH_LBM



import unittest
import aviary.api as av


BatteryBuilder = av.TestSubsystemBuilderBase.import_builder(
    'battery.battery_builder.BatteryBuilder')


@av.skipIfMissingDependencies(BatteryBuilder)
class TestBattery(av.TestSubsystemBuilderBase):

    def setUp(self):
        self.subsystem_builder = BatteryBuilder()
        self.aviary_values = av.AviaryValues()


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials

from aviary.subsystems.aerodynamics.flops_based.lift_dependent_drag import \
    LiftDependentDrag
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class LiftDependentDragTest(unittest.TestCase):

    def test_derivs_edge_interp(self):

        # Pressure in lbf/in**2 at 41000 ft.
        P = 2.60239151
        Sref = 1370.0

        CL = np.array([0.3, 0.35, 0.4, 0.45, 0.5, 0.55])
        mach = np.array([0.4, 0.45, 0.5, 0.55, 0.6, 0.85])
        lift = 0.5 * CL * Sref * 1.4 * P * mach ** 2

        nn = len(CL)

        prob = om.Problem(model=om.Group())
        prob.model.add_subsystem('drag', LiftDependentDrag(num_nodes=nn), promotes=['*'])
        prob.setup(force_alloc_complex=True)

        prob.set_val(Dynamic.Mission.MACH, val=mach)
        prob.set_val(Dynamic.Mission.LIFT, val=lift)
        prob.set_val(Dynamic.Mission.STATIC_PRESSURE, val=P)
        prob.set_val(Aircraft.Wing.AREA, val=Sref)
        prob.set_val(Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN, val=1.0)
        prob.set_val(Aircraft.Wing.SWEEP, val=25.03)

        prob.set_val(Aircraft.Wing.ASPECT_RATIO, val=11.05)

        prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD, val=0.123)
        prob.set_val(Mission.Design.LIFT_COEFFICIENT, val=1.28)
        prob.set_val(Mission.Design.MACH, val=0.765)

        prob.run_model()

        derivs = prob.check_partials(out_stream=None, method="cs")

        # TODO: need to test outputs too
        assert_check_partials(derivs, atol=1e-12, rtol=1e-12)

    def test_derivs_inner_interp(self):

        # Pressure in lbf/in**2 at 41000 ft.
        P = 2.60239151
        Sref = 1370.0

        CL = np.array([0.3, 0.35, 0.4, 0.45, 0.5, 0.55])
        mach = np.array([0.4, 0.45, 0.5, 0.55, 0.6, 0.85])
        lift = 0.5 * CL * Sref * 1.4 * P * mach ** 2

        nn = len(CL)

        prob = om.Problem(model=om.Group())
        prob.model.add_subsystem('drag', LiftDependentDrag(num_nodes=nn), promotes=['*'])
        prob.setup(force_alloc_complex=True)

        prob.set_val(Dynamic.Mission.MACH, val=mach)
        prob.set_val(Dynamic.Mission.LIFT, val=lift)
        prob.set_val(Dynamic.Mission.STATIC_PRESSURE, val=P)
        prob.set_val(Aircraft.Wing.AREA, val=Sref)
        prob.set_val(Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN, val=1.0)
        prob.set_val(Aircraft.Wing.SWEEP, val=25.07)

        prob.set_val(Aircraft.Wing.ASPECT_RATIO, val=11.05 * 0.5)

        prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD, val=0.132)
        prob.set_val(Mission.Design.LIFT_COEFFICIENT, val=0.1234)
        prob.set_val(Mission.Design.MACH, val=0.4321)

        prob.run_model()

        derivs = prob.check_partials(out_stream=None, method="cs")

        # TODO: need to test outputs too
        assert_check_partials(derivs, atol=1e-12, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om
import aviary.api as av
from aviary.examples.external_subsystems.battery.battery_variables import Aircraft, Mission


ExtendedMetaData = av.CoreMetaData


##### BATTERY VALUES #####

av.add_meta_data(
    Aircraft.Battery.CURRENT_MAX,
    units="A",
    desc="Max current through the pack",
    default_value=10.,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.EFFICIENCY,
    units=None,
    desc="Battery efficiency (eta)",
    default_value=0.85,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.ENERGY_REQUIRED,
    units="kW*h",
    desc="Required battery energy",
    default_value=65.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.HEAT_CAPACITY,
    units="J/(kg*K)",
    desc="mass-averaged specific heat (cp)",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.MASS,
    units="kg",
    desc="Battery mass",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.N_PARALLEL,
    units=None,
    desc="Number of cells in parallel, based on power constraint",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.N_SERIES,
    units=None,
    desc="Number of cells in series",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.VOLTAGE,
    units="V",
    desc="Nominal bus voltage",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.VOLUME,
    units="inch**3",
    desc="Battery volume",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

##### CASE VALUES #####

av.add_meta_data(
    Aircraft.Battery.Case.HEAT_CAPACITY,
    units="J/(kg*K)",
    desc="Mass-averaged specific heat (cp)",
    default_value=921.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.Case.WEIGHT_FRAC,
    units=None,
    desc="Case weight per unit of cell weight",
    default_value=1.3,
    meta_data=ExtendedMetaData
)

##### CELL VALUES #####

av.add_meta_data(
    Aircraft.Battery.Cell.DISCHARGE_RATE,
    units="A",
    desc="Cell discharge rate",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.Cell.ENERGY_CAPACITY_MAX,
    units="A*h",
    desc="Maximum energy capacity of a single cell (Q_max)",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.Cell.HEAT_CAPACITY,
    units="J/(kg*K)",
    desc="Specific heat of a battery cell",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.Cell.MASS,
    units="kg",
    desc="Mass of a single cell",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.Cell.VOLTAGE_LOW,
    units="V",
    desc="Cell Voltage at low SOC, before drop-off",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Aircraft.Battery.Cell.VOLUME,
    units="inch**3",
    desc="Volume of a single cell",
    default_value=1.0,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Mission.Battery.CURRENT,
    units="A",
    desc="Current output from the battery",
    default_value=None,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Mission.Battery.HEAT_OUT,
    units="W",
    desc="Heat generated by the battery",
    default_value=None,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Mission.Battery.STATE_OF_CHARGE,
    units=None,
    desc="State of charge of the battery (SOC)",
    default_value=None,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Mission.Battery.STATE_OF_CHARGE_RATE,
    units="1/s",
    desc="Time rate of change of the battery state of charge",
    default_value=None,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Mission.Battery.TEMPERATURE,
    units="K",
    desc="Battery pack bulk temperature",
    default_value=None,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Mission.Battery.VOLTAGE,
    units="V",
    desc="Total battery voltage",
    default_value=None,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Mission.Battery.VOLTAGE_THEVENIN,
    units="V",
    desc="Thevenin (Polarization) voltage",
    default_value=None,
    meta_data=ExtendedMetaData
)

av.add_meta_data(
    Mission.Battery.VOLTAGE_THEVENIN_RATE,
    units="V/s",
    desc="Time rate of change of Thevenin voltage",
    default_value=None,
    meta_data=ExtendedMetaData
)



"""
Attributes
----------
default_units : dict
    Matches each EngineModelVariables entry with default units (str)
"""
from enum import Enum, auto

import aviary.constants as constants


class EngineModelVariables(Enum):
    '''
    Define constants that map to supported variable names in an engine model.
    '''
    MACH = auto()
    ALTITUDE = auto()
    THROTTLE = auto()
    HYBRID_THROTTLE = auto()
    THRUST = auto()
    GROSS_THRUST = auto()
    RAM_DRAG = auto()
    FUEL_FLOW = auto()
    ELECTRIC_POWER = auto()
    NOX_RATE = auto()
    TEMPERATURE_ENGINE_T4 = auto()
    # EXIT_AREA = auto()


default_units = {
    EngineModelVariables.MACH: 'unitless',
    EngineModelVariables.ALTITUDE: 'ft',
    EngineModelVariables.THROTTLE: 'unitless',
    EngineModelVariables.HYBRID_THROTTLE: 'unitless',
    EngineModelVariables.THRUST: 'lbf',
    EngineModelVariables.GROSS_THRUST: 'lbf',
    EngineModelVariables.RAM_DRAG: 'lbf',
    EngineModelVariables.FUEL_FLOW: 'lb/h',
    EngineModelVariables.ELECTRIC_POWER: 'kW',
    EngineModelVariables.NOX_RATE: 'lb/h',
    EngineModelVariables.TEMPERATURE_ENGINE_T4: 'degR'
    # EngineModelVariables.EXIT_AREA: 'ft**2',
}


def convert_geopotential_altitude(altitude):
    """
    Converts altitudes from geopotential to geometric altitude
    Assumes altitude is provided in feet.

    Parameters
    ----------
    altitude_list : <(float, list of floats)>
        geopotential altitudes (in ft) to be converted.

    Returns
    ----------
    altitude_list : <list of floats>
        geometric altitudes (ft).
    """
    try:
        iter(altitude)
    except TypeError:
        altitude = [altitude]

    g = constants.GRAV_METRIC_FLOPS
    radius_earth = constants.RADIUS_EARTH_METRIC
    CM1 = 0.99850  # Center of mass (Earth)? Unknown
    OC2 = 26.76566e-10  # Unknown
    GNS = 9.8236930  # grav_accel_at_surface_earth?

    for (i, alt) in enumerate(altitude):
        HFT = alt
        HO = HFT * .30480  # convert ft to m
        Z = (HFT + (4.37 * (10 ** -8)) * (HFT ** 2.00850)) * .30480

        DH = float('inf')

        while abs(DH) > 1.0:
            R = radius_earth + Z
            GN = GNS * (radius_earth / R) ** (CM1 + 1.0)
            H =\
                (R * GN * ((R / radius_earth)**CM1 - 1.0)
                    / CM1 - Z * (R - Z / 2.0) * OC2) / g

            DH = HO - H
            Z += DH

        alt = Z / .30480  # convert m to ft
        altitude[i] = alt

    return altitude


# class InstallationDragFlag(Enum):
#     '''
#     Define constants that map to supported options for scaling of installation drag.
#     '''
#     OFF = auto()
#     DELTA_MAX_NOZZLE_AREA = auto()
#     MAX_NOZZLE_AREA = auto()
#     REF_NOZZLE_EXIT_AREA = auto()



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.interface.default_phase_info.two_dof import aero
from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.mission.gasp_based.ode.unsteady_solved.unsteady_control_iter_group import \
    UnsteadyControlIterGroup
from aviary.mission.gasp_based.ode.unsteady_solved.unsteady_solved_flight_conditions import \
    UnsteadySolvedFlightConditions
from aviary.variable_info.enums import SpeedType
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class TestUnsteadyAlphaThrustIterGroup(unittest.TestCase):

    def _test_unsteady_alpha_thrust_iter_group(self, ground_roll=False):
        nn = 5

        p = om.Problem()

        # TODO: paramport
        param_port = ParamPort()

        p.model.add_subsystem("params", param_port, promotes=["*"])

        fc = UnsteadySolvedFlightConditions(num_nodes=nn,
                                            input_speed_type=SpeedType.TAS,
                                            ground_roll=ground_roll)

        p.model.add_subsystem("fc", subsys=fc,
                              promotes_inputs=["*"],
                              promotes_outputs=["*"])

        g = UnsteadyControlIterGroup(num_nodes=nn,
                                     ground_roll=ground_roll,
                                     clean=True,
                                     core_subsystems=[aero])

        p.model.add_subsystem("iter_group",
                              subsys=g,
                              promotes_inputs=["*"],
                              promotes_outputs=["*"])

        for key, data in param_port.param_data.items():
            p.model.set_input_defaults(key, **data)

        p.setup(force_alloc_complex=True)

        p.final_setup()

        p.set_val(Dynamic.Mission.SPEED_OF_SOUND, 968.076 * np.ones(nn), units="ft/s")
        p.set_val("rho", 0.000659904 * np.ones(nn), units="slug/ft**3")
        p.set_val("TAS", 487 * np.ones(nn), units="kn")
        p.set_val("mass", 170_000 * np.ones(nn), units="lbm")
        p.set_val("dTAS_dr", 0.0 * np.ones(nn), units="kn/NM")

        if not ground_roll:
            p.set_val(Dynamic.Mission.FLIGHT_PATH_ANGLE, 0.0 * np.ones(nn), units="rad")
            p.set_val("alpha", 4 * np.ones(nn), units="deg")
            p.set_val("dh_dr", 0.0 * np.ones(nn), units=None)
            p.set_val("d2h_dr2", 0.0 * np.ones(nn), units="1/NM")

        p.set_val("thrust_req", 8000 * np.ones(nn), units="lbf")

        p.run_model()

        drag = p.model.get_val(Dynamic.Mission.DRAG, units="lbf")
        lift = p.model.get_val(Dynamic.Mission.LIFT, units="lbf")
        thrust_req = p.model.get_val("thrust_req", units="lbf")
        gamma = 0 if ground_roll else p.model.get_val(
            Dynamic.Mission.FLIGHT_PATH_ANGLE, units="deg")
        weight = p.model.get_val("mass", units="lbm") * GRAV_ENGLISH_LBM
        iwing = p.model.get_val(Aircraft.Wing.INCIDENCE, units="deg")
        alpha = iwing if ground_roll else p.model.get_val("alpha", units="deg")

        c_alphai = np.cos(np.radians(alpha - iwing))
        s_alphai = np.sin(np.radians(alpha - iwing))

        c_gamma = np.cos(np.radians(gamma))
        s_gamma = np.sin(np.radians(gamma))

        # 1. Test that forces balance along the velocity axis
        assert_near_equal(drag + thrust_req * s_gamma, thrust_req * c_alphai)

        # 2. Test that forces balance normal to the velocity axis
        assert_near_equal(lift + thrust_req * s_alphai, weight * c_gamma)

    def test_iter_group(self):
        for ground_roll in [False]:
            with self.subTest(msg=f"ground_roll={ground_roll}"):
                self._test_unsteady_alpha_thrust_iter_group(ground_roll=ground_roll)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class SimpleWingBendingFact(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Wing.AREA, val=0.0)

        add_aviary_input(self, Aircraft.Wing.SPAN, val=0.0)

        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, val=0.0)

        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD, val=0.0)

        add_aviary_input(self, Aircraft.Wing.STRUT_BRACING_FACTOR, val=0.0)

        add_aviary_input(self, Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR, val=0.0)

        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, val=0.0)

        add_aviary_input(self, Aircraft.Wing.SWEEP, val=0.0)

        add_aviary_output(self, Aircraft.Wing.BENDING_FACTOR, val=0.0)

        add_aviary_output(self, Aircraft.Wing.ENG_POD_INERTIA_FACTOR, val=0.0)

    def setup_partials(self):
        self.declare_partials(of=Aircraft.Wing.BENDING_FACTOR,
                              wrt=[Aircraft.Wing.STRUT_BRACING_FACTOR,
                                   Aircraft.Wing.SPAN,
                                   Aircraft.Wing.TAPER_RATIO,
                                   Aircraft.Wing.AREA,
                                   Aircraft.Wing.THICKNESS_TO_CHORD,
                                   Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR,
                                   Aircraft.Wing.ASPECT_RATIO,
                                   Aircraft.Wing.SWEEP])

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_wing_eng = aviary_options.get_val(Aircraft.Propulsion.TOTAL_NUM_WING_ENGINES)
        fstrt = inputs[Aircraft.Wing.STRUT_BRACING_FACTOR]
        span = inputs[Aircraft.Wing.SPAN]
        tr = inputs[Aircraft.Wing.TAPER_RATIO]
        area = inputs[Aircraft.Wing.AREA]
        tca = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]
        faert = inputs[Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR]
        ar = inputs[Aircraft.Wing.ASPECT_RATIO]
        sweep = inputs[Aircraft.Wing.SWEEP]

        C4 = 1.0 - 0.5 * faert
        C6 = 0.5 * faert - 0.16 * fstrt

        if ar <= 5.0:
            caya = 0.0
        else:
            caya = ar - 5.0

        tlam = np.tan(np.pi / 180. * sweep) - 2 * (1 - tr) / (ar * (1 + tr))

        slam = tlam / (1.0 + tlam**2)**0.5

        cayl = (1.0 - slam**2) * (1.0 + C6 * slam**2 + 0.03 * caya * C4 * slam)

        ems = 1.0 - 0.25 * fstrt

        outputs[Aircraft.Wing.BENDING_FACTOR] = \
            0.215 * (0.37 + 0.7 * tr) * (span**2 / area)**ems / (cayl * tca)

        outputs[Aircraft.Wing.ENG_POD_INERTIA_FACTOR] = 1.0 - 0.03 * num_wing_eng

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_wing_eng = aviary_options.get_val(Aircraft.Propulsion.TOTAL_NUM_WING_ENGINES)
        fstrt = inputs[Aircraft.Wing.STRUT_BRACING_FACTOR]
        span = inputs[Aircraft.Wing.SPAN]
        tr = inputs[Aircraft.Wing.TAPER_RATIO]
        area = inputs[Aircraft.Wing.AREA]
        tca = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]
        faert = inputs[Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR]
        ar = inputs[Aircraft.Wing.ASPECT_RATIO]
        sweep = inputs[Aircraft.Wing.SWEEP]

        C4 = 1.0 - 0.5 * faert
        C6 = 0.5 * faert - 0.16 * fstrt

        if ar <= 5.0:
            caya = 0.0
            dcayl_ar = 0.0
        else:
            caya = ar - 5.0

        deg2rad = np.pi / 180.
        den = ar * (1 + tr)
        tlam = np.tan(deg2rad * sweep) - 2 * (1 - tr) / den
        dtlam_sweep = deg2rad / np.cos(deg2rad * sweep) ** 2
        dtlam_ar = 2.0 * (1 - tr) / (ar * den)
        dtlam_tr = 2.0 * (1.0 / den + (1 - tr) / den**2 * ar)

        den = 1.0 + tlam**2
        slam = tlam / den**0.5
        dslam = 1.0 / den**0.5 - tlam**2 / den**1.5

        term1 = 1.0 - slam**2
        term2 = 1.0 + C6 * slam**2 + 0.03 * caya * C4 * slam
        cayl = term1 * term2
        dcayl_slam = -2.0 * slam * term2 + term1 * (2.0 * slam * C6 + 0.03 * caya * C4)
        dcayl_faert = term1 * slam * 0.5 * (slam - 0.03 * caya)
        dcayl_fstrt = -term1 * 0.16 * slam**2
        if ar > 5.0:
            dcayl_ar = term1 * 0.03 * C4 * slam

        ems = 1.0 - 0.25 * fstrt

        # bend = 0.215 * (0.37 + 0.7 * tr) * (span**2 / area)**ems / (cayl * tca)
        term1 = (0.37 + 0.7 * tr)
        term2a = (span**2 / area)
        term2 = term2a ** ems
        term3 = 1.0 / (cayl * tca)
        dbend_exp = -0.215 * term1 * term2 * term3 * np.log(term2a) * 0.25
        dbend_tr = 0.215 * 0.7 * term2 * term3
        dbend_cayl = -0.215 * term1 * term2 * tca * term3**2

        J[Aircraft.Wing.BENDING_FACTOR, Aircraft.Wing.STRUT_BRACING_FACTOR] = \
            dbend_exp + dbend_cayl * dcayl_fstrt

        J[Aircraft.Wing.BENDING_FACTOR, Aircraft.Wing.SPAN] = \
            2.0 * 0.215 * term1 * ems * term2a**(ems - 1) * term3 * span / area

        J[Aircraft.Wing.BENDING_FACTOR, Aircraft.Wing.TAPER_RATIO] = \
            dbend_tr + dbend_cayl * (dcayl_slam * dslam * dtlam_tr)

        J[Aircraft.Wing.BENDING_FACTOR, Aircraft.Wing.AREA] = \
            -0.215 * term1 * ems * term2a**(ems - 1) * term3 * term2a / area

        J[Aircraft.Wing.BENDING_FACTOR, Aircraft.Wing.THICKNESS_TO_CHORD] = \
            -0.215 * term1 * term2 * cayl * term3**2

        J[Aircraft.Wing.BENDING_FACTOR, Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR] = \
            dbend_cayl * dcayl_faert

        J[Aircraft.Wing.BENDING_FACTOR, Aircraft.Wing.ASPECT_RATIO] = \
            dbend_cayl * (dcayl_ar + dcayl_slam * dslam * dtlam_ar)

        J[Aircraft.Wing.BENDING_FACTOR, Aircraft.Wing.SWEEP] = \
            dbend_cayl * (dcayl_slam * dslam * dtlam_sweep)



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.propulsion.engine_deck import EngineDeck
from aviary.subsystems.propulsion.engine_scaling import EngineScaling
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.preprocessors import preprocess_propulsion
from aviary.utils.functions import get_path
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class EngineScalingTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem(model=om.Group())

    def test_case(self):
        nn = 4
        count = 1

        filename = 'models/engines/turbofan_28k.deck'
        filename = get_path(filename)

        options = AviaryValues()
        options.set_val(Aircraft.Engine.DATA_FILE, filename)
        options.set_val(Aircraft.Engine.SUBSONIC_FUEL_FLOW_SCALER, 0.9)
        # make supersonic scaling factor extremely high so it is obvious if it gets used
        options.set_val(Aircraft.Engine.SUPERSONIC_FUEL_FLOW_SCALER, 100)
        options.set_val(Aircraft.Engine.FUEL_FLOW_SCALER_CONSTANT_TERM, 1.15)
        options.set_val(Aircraft.Engine.FUEL_FLOW_SCALER_LINEAR_TERM, 1.05)
        options.set_val(Aircraft.Engine.CONSTANT_FUEL_CONSUMPTION, 10.0, units='lbm/h')
        options.set_val(Aircraft.Engine.SCALE_PERFORMANCE, True)
        options.set_val(Aircraft.Engine.SCALE_FACTOR, 0.9)
        options.set_val(Aircraft.Engine.GENERATE_FLIGHT_IDLE, True)
        options.set_val(Aircraft.Engine.IGNORE_NEGATIVE_THRUST, False)
        options.set_val(Aircraft.Engine.FLIGHT_IDLE_THRUST_FRACTION, 0.0)
        options.set_val(Aircraft.Engine.FLIGHT_IDLE_MAX_FRACTION, 1.0)
        options.set_val(Aircraft.Engine.FLIGHT_IDLE_MIN_FRACTION, 0.08)
        options.set_val(Aircraft.Engine.GEOPOTENTIAL_ALT, False)
        options.set_val(Aircraft.Engine.INTERPOLATION_METHOD, 'slinear')

        # engine1 uses all scaling factors
        engine1 = EngineDeck(options=options)

        preprocess_propulsion(options, [engine1])

        options.set_val(Mission.Summary.FUEL_FLOW_SCALER, 10.)

        self.prob.model.add_subsystem('engine', EngineScaling(
            num_nodes=nn,
            aviary_options=options),
            promotes=['*'])
        self.prob.setup(force_alloc_complex=True)
        self.prob.set_val('thrust_net_unscaled', np.ones(
            [nn, count]) * 1000, units='lbf')
        self.prob.set_val('fuel_flow_rate_unscaled', np.ones(
            [nn, count]) * 100, units='lbm/h')
        self.prob.set_val('nox_rate_unscaled', np.ones([nn, count]) * 10, units='lbm/h')
        self.prob.set_val(Dynamic.Mission.MACH, np.linspace(
            0, 0.75, nn), units='unitless')
        self.prob.set_val(Aircraft.Engine.SCALE_FACTOR,
                          options.get_val(Aircraft.Engine.SCALE_FACTOR))

        self.prob.run_model()

        thrust = self.prob.get_val(Dynamic.Mission.THRUST)
        fuel_flow = self.prob.get_val(Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE)
        nox_rate = self.prob.get_val(Dynamic.Mission.NOX_RATE)
        # exit_area = self.prob.get_val(Dynamic.Mission.EXIT_AREA)

        thrust_expected = np.array([900., 900., 900., 900])

        fuel_flow_expected = np.array([-1836.55, -1836.55, -1836.55, -1836.55])

        nox_rate_expected = np.array([9., 9., 9., 9])

        assert_near_equal(thrust, thrust_expected, tolerance=1e-10)
        assert_near_equal(fuel_flow, fuel_flow_expected, tolerance=1e-10)
        assert_near_equal(nox_rate, nox_rate_expected, tolerance=1e-10)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-11, rtol=1e-10)


if __name__ == "__main__":
    unittest.main()



'''
Define utilities to calculate the estimated mass of the crew (both flight and
non-flight) as well as their baggage.
'''
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class NonFlightCrewMass(om.ExplicitComponent):
    '''
    Calculate the estimated mass for the non-flight and their baggage.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(
            self, Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS_SCALER, 1.)

        add_aviary_output(self, Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS, 0)

    def setup_partials(self):
        self.declare_partials(
            Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS,
            Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS_SCALER
        )

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        aviary_options: AviaryValues = self.options['aviary_options']
        flight_attendants_count = \
            aviary_options.get_val(Aircraft.CrewPayload.NUM_FLIGHT_ATTENDANTS)
        galley_crew_count = \
            aviary_options.get_val(Aircraft.CrewPayload.NUM_GALLEY_CREW)

        mass_per_flight_attendant = self._mass_per_flight_attendant
        mass_per_galley_crew = self._mass_per_galley_crew

        mass_scaler = \
            inputs[Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS_SCALER]

        outputs[Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS] = (
            flight_attendants_count * mass_per_flight_attendant
            + galley_crew_count * mass_per_galley_crew
        ) * mass_scaler

    def compute_partials(self, inputs, J, discrete_inputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        flight_attendants_count = \
            aviary_options.get_val(Aircraft.CrewPayload.NUM_FLIGHT_ATTENDANTS)
        galley_crew_count = \
            aviary_options.get_val(Aircraft.CrewPayload.NUM_GALLEY_CREW)

        mass_per_flight_attendant = self._mass_per_flight_attendant
        mass_per_galley_crew = self._mass_per_galley_crew

        J[
            Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS,
            Aircraft.CrewPayload.NON_FLIGHT_CREW_MASS_SCALER
        ] = (
            flight_attendants_count * mass_per_flight_attendant
            + galley_crew_count * mass_per_galley_crew
        )

    _mass_per_flight_attendant = 155.  # lbm
    _mass_per_galley_crew = 200.  # lbm


class FlightCrewMass(om.ExplicitComponent):
    '''
    Calculate the estimated mass for the flight crew and their baggage.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(
            self, Aircraft.CrewPayload.FLIGHT_CREW_MASS_SCALER, 1.)

        add_aviary_output(self, Aircraft.CrewPayload.FLIGHT_CREW_MASS, 0.)

    def setup_partials(self):
        self.declare_partials(
            Aircraft.CrewPayload.FLIGHT_CREW_MASS,
            Aircraft.CrewPayload.FLIGHT_CREW_MASS_SCALER
        )

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        aviary_options: AviaryValues = self.options['aviary_options']
        flight_crew_count = \
            aviary_options.get_val(Aircraft.CrewPayload.NUM_FLIGHT_CREW)

        mass_per_flight_crew = self._mass_per_flight_crew(inputs)

        mass_scaler = inputs[Aircraft.CrewPayload.FLIGHT_CREW_MASS_SCALER]

        outputs[Aircraft.CrewPayload.FLIGHT_CREW_MASS] = \
            flight_crew_count * mass_per_flight_crew * mass_scaler

    def compute_partials(self, inputs, J, discrete_inputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        flight_crew_count = \
            aviary_options.get_val(Aircraft.CrewPayload.NUM_FLIGHT_CREW)

        mass_per_flight_crew = self._mass_per_flight_crew(inputs)

        J[
            Aircraft.CrewPayload.FLIGHT_CREW_MASS,
            Aircraft.CrewPayload.FLIGHT_CREW_MASS_SCALER
        ] = flight_crew_count * mass_per_flight_crew

    def _mass_per_flight_crew(self, inputs):
        '''
        Return the mass, in pounds, of one member of the flight crew and
        their baggage.
        '''
        aviary_options: AviaryValues = self.options['aviary_options']
        mass_per_flight_crew = 225.0  # lbm

        # account for machine precision error
        if 0.9 <= aviary_options.get_val(Aircraft.LandingGear.CARRIER_BASED):
            mass_per_flight_crew -= 35.0  # lbm

        return mass_per_flight_crew



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.geometry.gasp_based.electric import CableSize
from aviary.variable_info.variables import Aircraft


class ElectricTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem("cable", CableSize(), promotes=["*"])

        self.prob.model.set_input_defaults(
            Aircraft.Engine.WING_LOCATIONS, 0.35, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, 128, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, 10, units="ft"
        )  # not actual GASP value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Electrical.HYBRID_CABLE_LENGTH], 64.8, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-8, rtol=1e-8)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import nacelle_count_factor
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class ThrustReverserMass(om.ExplicitComponent):
    '''
    Calculates mass of thrust reversers for entire set of an engine model.
    The methodology is based on the FLOPS weight equations, modified to
    output mass instead of weight.

    Assumptions
    -----------
    Total propulsion level thrust is the same as engine set's thrust

    Currently assumed engine set level thrust reversers is the same
    as propulsion-level total sum of thrust reverser masses
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Engine.THRUST_REVERSERS_MASS_SCALER, val=0.0)
        add_aviary_input(self, Aircraft.Engine.SCALED_SLS_THRUST, val=np.zeros(1))

        add_aviary_output(
            self, Aircraft.Engine.THRUST_REVERSERS_MASS, val=np.zeros(1))
        add_aviary_output(
            self, Aircraft.Propulsion.TOTAL_THRUST_REVERSERS_MASS, val=0)

        self.declare_partials('*', '*')

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_eng = aviary_options.get_val(Aircraft.Engine.NUM_ENGINES)
        scaler = inputs[Aircraft.Engine.THRUST_REVERSERS_MASS_SCALER]
        max_thrust = inputs[Aircraft.Engine.SCALED_SLS_THRUST]
        nac_count = nacelle_count_factor(num_eng)

        thrust_reverser_mass = .034 * max_thrust * nac_count * scaler / GRAV_ENGLISH_LBM
        outputs[Aircraft.Engine.THRUST_REVERSERS_MASS] = thrust_reverser_mass
        outputs[Aircraft.Propulsion.TOTAL_THRUST_REVERSERS_MASS] = sum(
            thrust_reverser_mass)

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_eng = aviary_options.get_val(Aircraft.Engine.NUM_ENGINES)
        scaler = inputs[Aircraft.Engine.THRUST_REVERSERS_MASS_SCALER]
        max_thrust = inputs[Aircraft.Engine.SCALED_SLS_THRUST]
        nac_count = nacelle_count_factor(num_eng)

        J[Aircraft.Engine.THRUST_REVERSERS_MASS,
            Aircraft.Engine.THRUST_REVERSERS_MASS_SCALER] = \
            .034 * max_thrust * nac_count / GRAV_ENGLISH_LBM
        J[Aircraft.Engine.THRUST_REVERSERS_MASS,
            Aircraft.Engine.SCALED_SLS_THRUST] = .034*nac_count * scaler / GRAV_ENGLISH_LBM

        J[Aircraft.Propulsion.TOTAL_THRUST_REVERSERS_MASS,
            Aircraft.Engine.THRUST_REVERSERS_MASS_SCALER] = \
            .034 * max_thrust * nac_count / GRAV_ENGLISH_LBM
        J[Aircraft.Propulsion.TOTAL_THRUST_REVERSERS_MASS,
            Aircraft.Engine.SCALED_SLS_THRUST] = .034*nac_count * scaler / GRAV_ENGLISH_LBM



import numpy as np
import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.mission.gasp_based.flight_conditions import FlightConditions
from aviary.mission.gasp_based.ode.base_ode import BaseODE
from aviary.mission.gasp_based.ode.climb_eom import ClimbRates
from aviary.mission.gasp_based.ode.constraints.flight_constraints import \
    FlightConstraints
from aviary.mission.gasp_based.ode.constraints.speed_constraints import SpeedConstraints
from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.subsystems.aerodynamics.aerodynamics_builder import AerodynamicsBuilderBase
from aviary.subsystems.propulsion.propulsion_builder import PropulsionBuilderBase
from aviary.variable_info.enums import AnalysisScheme, AlphaModes, SpeedType
from aviary.variable_info.variables import Dynamic
from aviary.mission.ode.specific_energy_rate import SpecificEnergyRate
from aviary.mission.ode.altitude_rate import AltitudeRate


class ClimbODE(BaseODE):
    """ODE for quasi-steady climb.

    This ODE has a ``KSComp`` which allows for the switching of obeying an EAS
    constraint or a cruise Mach constraint, whichever is being violated.
    """

    def initialize(self):
        super().initialize()
        self.options.declare("input_speed_type", default=SpeedType.EAS, types=SpeedType,
                             desc="Whether the speed is given as a equivalent airspeed, true airspeed, or mach number")
        self.options.declare("alt_trigger_units", default='ft',
                             desc='The units that the altitude trigger is provided in')
        self.options.declare("speed_trigger_units", default='kn',
                             desc='The units that the speed trigger is provided in.')
        self.options.declare("input_speed_type", default=SpeedType.EAS, types=SpeedType,
                             desc="Whether the speed is given as a equivalent airspeed, true airspeed, or mach number")
        self.options.declare("EAS_target", desc="target climbing EAS in knots")
        self.options.declare(
            "mach_cruise", default=0, desc="targeted cruise mach number"
        )

    def setup(self):
        nn = self.options["num_nodes"]
        analysis_scheme = self.options["analysis_scheme"]
        aviary_options = self.options['aviary_options']
        core_subsystems = self.options['core_subsystems']
        input_speed_type = self.options["input_speed_type"]

        if input_speed_type is SpeedType.EAS:
            speed_inputs = ["EAS"]
            speed_outputs = ["mach", ("TAS", Dynamic.Mission.VELOCITY)]
        elif input_speed_type is SpeedType.MACH:
            speed_inputs = ["mach"]
            speed_outputs = ["EAS", ("TAS", Dynamic.Mission.VELOCITY)]

        # TODO: paramport
        self.add_subsystem("params", ParamPort(), promotes=["*"])

        self.add_subsystem(
            "USatm",
            USatm1976Comp(
                num_nodes=nn),
            promotes_inputs=[
                ("h",
                 Dynamic.Mission.ALTITUDE)],
            promotes_outputs=[
                "rho",
                ("sos",
                 Dynamic.Mission.SPEED_OF_SOUND),
                ("temp",
                 Dynamic.Mission.TEMPERATURE),
                ("pres",
                 Dynamic.Mission.STATIC_PRESSURE),
                "viscosity"],
        )

        if analysis_scheme is AnalysisScheme.COLLOCATION:
            EAS_target = self.options["EAS_target"]
            mach_cruise = self.options["mach_cruise"]

            constraint_args = {}
            integration_states = []
            constraint_inputs = []

            mach_balance_group = self.add_subsystem(
                "mach_balance_group", subsys=om.Group(), promotes=["*"]
            )

            mach_balance_group.nonlinear_solver = om.NewtonSolver()
            mach_balance_group.nonlinear_solver.options["solve_subsystems"] = True
            mach_balance_group.nonlinear_solver.options["iprint"] = 0
            mach_balance_group.nonlinear_solver.options["atol"] = 1e-7
            mach_balance_group.nonlinear_solver.options["rtol"] = 1e-7
            mach_balance_group.nonlinear_solver.linesearch = om.BoundsEnforceLS()
            mach_balance_group.linear_solver = om.DirectSolver(assemble_jac=True)
            mach_balance_group.add_subsystem(
                "speeds",
                SpeedConstraints(
                    num_nodes=nn, EAS_target=EAS_target, mach_cruise=mach_cruise
                ),
                promotes_inputs=["EAS", Dynamic.Mission.MACH],
                promotes_outputs=["speed_constraint"],
            )
            mach_balance_group.add_subsystem(
                "ks",
                om.KSComp(width=2, vec_size=nn, units="unitless"),
                promotes_inputs=[("g", "speed_constraint")],
                promotes_outputs=["KS"],
            )
            speed_bal = om.BalanceComp(
                name="EAS",
                val=EAS_target * np.ones(nn),
                units="kn",
                lhs_name="KS",
                rhs_val=0.0,
                eq_units='unitless',
                upper=350,
                lower=0,
            )
            mach_balance_group.add_subsystem(
                "speed_bal",
                speed_bal,
                promotes_inputs=["KS"],
                promotes_outputs=["EAS"],
            )

            lift_balance_group = self.add_subsystem(
                "lift_balance_group", subsys=om.Group(), promotes=["*"]
            )
            flight_condition_group = mach_balance_group

        elif analysis_scheme is AnalysisScheme.SHOOTING:
            constraint_args = {'analysis_scheme': AnalysisScheme.SHOOTING,
                               'alt_trigger_units': self.options["alt_trigger_units"],
                               'speed_trigger_units': self.options["speed_trigger_units"]}

            integration_states = ["t_curr", Dynamic.Mission.DISTANCE]
            constraint_inputs = ["alt_trigger", "speed_trigger"]

            lift_balance_group = self
            flight_condition_group = self

        flight_condition_group.add_subsystem(
            "fc",
            FlightConditions(num_nodes=nn, input_speed_type=input_speed_type),
            promotes_inputs=["rho", Dynamic.Mission.SPEED_OF_SOUND] + speed_inputs,
            promotes_outputs=[Dynamic.Mission.DYNAMIC_PRESSURE,] + speed_outputs,
        )

        kwargs = {'num_nodes': nn, 'aviary_inputs': aviary_options,
                  'method': 'cruise'}
        # collect the propulsion group names for later use with
        prop_groups = []
        for subsystem in core_subsystems:
            system = subsystem.build_mission(**kwargs)
            if system is not None:
                if isinstance(subsystem, AerodynamicsBuilderBase):
                    lift_balance_group.add_subsystem(subsystem.name,
                                                     system,
                                                     promotes_inputs=subsystem.mission_inputs(
                                                         **kwargs),
                                                     promotes_outputs=subsystem.mission_outputs(**kwargs))
                else:
                    if isinstance(subsystem, PropulsionBuilderBase):
                        prop_groups.append(subsystem.name)

                    self.add_subsystem(subsystem.name,
                                       system,
                                       promotes_inputs=subsystem.mission_inputs(
                                           **kwargs),
                                       promotes_outputs=subsystem.mission_outputs(**kwargs))

        # maybe replace this with the solver in AddAlphaControl?
        lift_balance_group.nonlinear_solver = om.NewtonSolver()
        lift_balance_group.nonlinear_solver.options["solve_subsystems"] = True
        lift_balance_group.nonlinear_solver.options["iprint"] = 0
        lift_balance_group.nonlinear_solver.options["atol"] = 1e-7
        lift_balance_group.nonlinear_solver.options["rtol"] = 1e-7
        lift_balance_group.nonlinear_solver.linesearch = om.BoundsEnforceLS()
        lift_balance_group.linear_solver = om.DirectSolver(assemble_jac=True)

        lift_balance_group.add_subsystem(
            "eom",
            ClimbRates(
                num_nodes=nn,
                analysis_scheme=analysis_scheme),
            promotes_inputs=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.VELOCITY,
                Dynamic.Mission.DRAG,
                Dynamic.Mission.THRUST_TOTAL,] +
            integration_states,
            promotes_outputs=[
                Dynamic.Mission.ALTITUDE_RATE,
                Dynamic.Mission.DISTANCE_RATE,
                "required_lift",
                Dynamic.Mission.FLIGHT_PATH_ANGLE,
            ],
        )

        self.AddAlphaControl(
            alpha_group=lift_balance_group,
            alpha_mode=AlphaModes.REQUIRED_LIFT,
            add_default_solver=False,
            num_nodes=nn)

        self.add_subsystem(
            "constraints",
            FlightConstraints(num_nodes=nn, **constraint_args),
            promotes_inputs=[
                "alpha",
                "rho",
                "CL_max",
                Dynamic.Mission.FLIGHT_PATH_ANGLE,
                Dynamic.Mission.MASS,
                ("TAS", Dynamic.Mission.VELOCITY),
            ]
            + ["aircraft:*"]
            + constraint_inputs,
            promotes_outputs=["theta", "TAS_violation"],
        )

        # the last two subsystems will also be used for constraints
        self.add_excess_rate_comps(nn)

        if analysis_scheme is AnalysisScheme.COLLOCATION:
            self.set_order(['params', 'USatm', 'mach_balance_group'] + prop_groups +
                           ['lift_balance_group', 'constraints', 'SPECIFIC_ENERGY_RATE_EXCESS', 'ALTITUDE_RATE_MAX'])

        ParamPort.set_default_vals(self)
        self.set_input_defaults("CL_max", val=5 * np.ones(nn), units="unitless")
        self.set_input_defaults(Dynamic.Mission.ALTITUDE,
                                val=500 * np.ones(nn), units='ft')
        self.set_input_defaults(Dynamic.Mission.MASS,
                                val=174000 * np.ones(nn), units='lbm')
        self.set_input_defaults(Dynamic.Mission.MACH,
                                val=0 * np.ones(nn), units="unitless")



import unittest
import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.assert_utils import assert_check_partials

from aviary.subsystems.aerodynamics.flops_based.buffet_lift import BuffetLift
from aviary.variable_info.variables import Aircraft, Mission


class TestBuffetLift(unittest.TestCase):

    def test_derivs(self):
        mach = [.2, .25, .3, .35, .4, .45, .5, .55, .6, .7, .75, .775, .8, .825, .85, .875,
                0.9, 0.95, 1.05, 1.1]
        nn = len(mach)

        prob = om.Problem()
        model = prob.model

        model.add_subsystem('BUF', BuffetLift(num_nodes=nn),
                            promotes_inputs=[Aircraft.Wing.ASPECT_RATIO,
                                             Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN,
                                             Aircraft.Wing.SWEEP,
                                             Aircraft.Wing.THICKNESS_TO_CHORD,
                                             Mission.Design.MACH])

        prob.setup(force_alloc_complex=True)

        prob.set_val('BUF.mach', mach)

        prob.set_val(Aircraft.Wing.THICKNESS_TO_CHORD, 0.13)  # 0.1632
        prob.set_val(Aircraft.Wing.SWEEP, 25)
        prob.set_val(Aircraft.Wing.ASPECT_RATIO, 11.232936003236246)
        prob.set_val(Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN, 0.0)
        prob.set_val(Mission.Design.MACH, 0.801)

        prob.run_model()
        derivs = prob.check_partials(compact_print=True, method="cs")

        assert_check_partials(derivs, atol=1e-12, rtol=1e-12)

        truth = np.array([0.86693558,  0.83439662,  0.79510222,  0.74915741,  0.6965622,
                          0.63256819,  0.57123245,  0.51727722,  0.45729095,  0.31903298,
                          0.2212182,  0.15774801,  0.08446832, -0.00190504, -0.09177745,
                          -0.18507041, -0.28122907, -0.4820969, -0.91803465, -1.15310457])

        assert_near_equal(prob['BUF.DELCLB'], truth, 1e-7)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class SizeEngine(om.ExplicitComponent):
    '''
    Calculates thrust scaling factors for dynamic mission parameters. Designed for use
    with EngineDecks.

    Can be vectorized for all unique engines present on aircraft. Each index represents a
    single instance of an engine model.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        # count = len(self.options['aviary_options'].get_val('engine_models'))
        count = 1

        add_aviary_input(self, Aircraft.Engine.SCALED_SLS_THRUST, val=np.zeros(count))

        add_aviary_output(self, Aircraft.Engine.SCALE_FACTOR, val=np.zeros(count))

        # variables that also may require scaling
        # TODO - inlet_weight <input>
        # TODO - inlet_weight_scaling_exp <option>
        # TODO - nozzle_weight
        # TODO - nacelle_avg_length: function of sqrt(scale_factor)
        # TODO - nacelle_avg_diam: function of sqrt(scale_factor)
        # TODO - nacelle_wetted_area: if length, diam get scaled - this should be covered by geom

    def compute(self, inputs, outputs):
        options: AviaryValues = self.options['aviary_options']
        # engine_models = options.get_val('engine_models')
        scale_engine = options.get_val(Aircraft.Engine.SCALE_PERFORMANCE)

        reference_sls_thrust = options.get_val(
            Aircraft.Engine.REFERENCE_SLS_THRUST, units='lbf')

        scaled_sls_thrust = inputs[Aircraft.Engine.SCALED_SLS_THRUST]

        # set a default scaling factor of 1 for each engine
        # nm = len(engine_models)

        # use dtype to make complex safe
        # engine_scale_factor = np.ones(nm, dtype=scaled_sls_thrust.dtype)

        # Engine is only scaled if required
        # engine scale factor is ratio of scaled thrust target and reference thrust
        # scale_idx = np.where(scale_engine)
        # engine_scale_factor[scale_idx] = scaled_sls_thrust[scale_idx] / \
        #     reference_sls_thrust[scale_idx]
        engine_scale_factor = 1
        if scale_engine:
            engine_scale_factor = scaled_sls_thrust / reference_sls_thrust

        outputs[Aircraft.Engine.SCALE_FACTOR] = engine_scale_factor

    def setup_partials(self):
        # count = len(self.options['aviary_options'].get_val('engine_models'))

        # shape = np.arange(count, dtype=int)

        self.declare_partials(Aircraft.Engine.SCALE_FACTOR,
                              Aircraft.Engine.SCALED_SLS_THRUST)
        #   rows=shape, cols=shape)

    def compute_partials(self, inputs, J):
        options: AviaryValues = self.options['aviary_options']
        # engine_models = options.get_val('engine_models')
        scale_engine = options.get_val(Aircraft.Engine.SCALE_PERFORMANCE)
        reference_sls_thrust = options.get_val(
            Aircraft.Engine.REFERENCE_SLS_THRUST, units='lbf')

        # nm = len(engine_models)

        scaled_sls_thrust = inputs[Aircraft.Engine.SCALED_SLS_THRUST]
        # use dtype to make complex safe
        # deriv_scale_factor = np.zeros(nm, dtype=scaled_sls_thrust.dtype)

        # scale_idx = np.where(scale_engine)
        # deriv_scale_factor[scale_idx] = 1.0 / reference_sls_thrust[scale_idx]

        deriv_scale_factor = 0
        if scale_engine:
            deriv_scale_factor = 1.0 / reference_sls_thrust

        J[Aircraft.Engine.SCALE_FACTOR,
            Aircraft.Engine.SCALED_SLS_THRUST] = deriv_scale_factor



import unittest
import aviary.api as av


path_to_builder = 'OAS_weight.OAS_wing_weight_builder.OASWingWeightBuilder'
OASWingWeightBuilder = av.TestSubsystemBuilderBase.import_builder(path_to_builder)


@av.skipIfMissingDependencies(OASWingWeightBuilder)
class TestStructures(av.TestSubsystemBuilderBase):

    def setUp(self):
        self.subsystem_builder = OASWingWeightBuilder()
        self.aviary_values = av.AviaryValues()


if __name__ == '__main__':
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.variable_info.variables import Dynamic


class RangeRate(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(
            Dynamic.Mission.ALTITUDE_RATE,
            val=np.ones(nn),
            desc='climb rate',
            units='m/s')
        self.add_input(
            Dynamic.Mission.VELOCITY,
            val=np.ones(nn),
            desc='current velocity',
            units='m/s')
        self.add_output(
            Dynamic.Mission.DISTANCE_RATE,
            val=np.ones(nn),
            desc='current horizontal velocity (assumed no wind)',
            units='m/s')

    def compute(self, inputs, outputs):
        climb_rate = inputs[Dynamic.Mission.ALTITUDE_RATE]
        velocity = inputs[Dynamic.Mission.VELOCITY]
        climb_rate_2 = climb_rate**2
        velocity_2 = velocity**2
        if (climb_rate_2 >= velocity_2).any():
            raise om.AnalysisError(
                "WARNING: climb rate exceeds velocity (range_rate.py)")
        outputs[Dynamic.Mission.DISTANCE_RATE] = (velocity_2 - climb_rate_2)**0.5

    def setup_partials(self):
        arange = np.arange(self.options['num_nodes'])
        self.declare_partials(
            Dynamic.Mission.DISTANCE_RATE, [
                Dynamic.Mission.ALTITUDE_RATE, Dynamic.Mission.VELOCITY], rows=arange, cols=arange)

    def compute_partials(self, inputs, J):
        climb_rate = inputs[Dynamic.Mission.ALTITUDE_RATE]
        velocity = inputs[Dynamic.Mission.VELOCITY]

        J[Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.ALTITUDE_RATE] = -climb_rate / \
            (velocity**2 - climb_rate**2)**0.5
        J[Dynamic.Mission.DISTANCE_RATE, Dynamic.Mission.VELOCITY] = velocity / \
            (velocity**2 - climb_rate**2)**0.5



import numpy as np
import openmdao.api as om

from aviary.variable_info.variables import Dynamic


class DynamicPressure(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

        self.options.declare(
            'gamma', default=1.4, desc='Ratio of specific heats for air.')

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(
            Dynamic.Mission.STATIC_PRESSURE, np.ones(nn), units='lbf/ft**2',
            desc='Static pressure at each evaulation point.')

        self.add_input(
            Dynamic.Mission.MACH, np.ones(nn), units='unitless',
            desc='Mach at each evaulation point.')

        self.add_output(
            Dynamic.Mission.DYNAMIC_PRESSURE, val=np.ones(nn), units='lbf/ft**2',
            desc='pressure caused by fluid motion')

    def setup_partials(self):
        nn = self.options['num_nodes']

        rows_cols = np.arange(nn)

        self.declare_partials(
            Dynamic.Mission.DYNAMIC_PRESSURE, [
                Dynamic.Mission.STATIC_PRESSURE, Dynamic.Mission.MACH],
            rows=rows_cols, cols=rows_cols)

    def compute(self, inputs, outputs):
        gamma = self.options['gamma']
        P = inputs[Dynamic.Mission.STATIC_PRESSURE]
        M = inputs[Dynamic.Mission.MACH]

        outputs[Dynamic.Mission.DYNAMIC_PRESSURE] = 0.5 * gamma * P * M**2

    def compute_partials(self, inputs, partials):
        gamma = self.options['gamma']
        P = inputs[Dynamic.Mission.STATIC_PRESSURE]
        M = inputs[Dynamic.Mission.MACH]

        partials[Dynamic.Mission.DYNAMIC_PRESSURE, Dynamic.Mission.MACH] = gamma * P * M
        partials[Dynamic.Mission.DYNAMIC_PRESSURE,
                 Dynamic.Mission.STATIC_PRESSURE] = 0.5 * gamma * M**2



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import \
    distributed_engine_count_factor
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission

# TODO: update non-transport components to new standard to remove these variables
_wing_engine_count_factor = 'aircraft:propulsion:control:wing_engine_count_factor'
_fuse_engine_count_factor = 'aircraft:propulsion:control:fuselage_engine_count_factor'
_max_mach = 'aircraft:design:dimensions:max_mach'


class TransportHydraulicsGroupMass(om.ExplicitComponent):
    '''
    Calculates the mass of the hydraulics group using the transport/general aviation
    method. The methodology is based on the FLOPS weight equations, modified to
    output mass instead of weight.

    ASSUMPTIONS: all engines use hydraulics which follow this equation
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuselage.PLANFORM_AREA, val=0.0)

        add_aviary_input(self, Aircraft.Hydraulics.SYSTEM_PRESSURE, val=0.0)

        add_aviary_input(self, Aircraft.Hydraulics.MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.Wing.AREA, val=0.0)

        add_aviary_input(self, Aircraft.Wing.VAR_SWEEP_MASS_PENALTY, val=0.0)

        add_aviary_output(self, Aircraft.Hydraulics.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_wing_eng = aviary_options.get_val(
            Aircraft.Propulsion.TOTAL_NUM_WING_ENGINES)
        num_fuse_eng = aviary_options.get_val(
            Aircraft.Propulsion.TOTAL_NUM_FUSELAGE_ENGINES)
        num_wing_eng_fact = distributed_engine_count_factor(num_wing_eng)
        num_fuse_eng_fact = distributed_engine_count_factor(num_fuse_eng)

        planform = inputs[Aircraft.Fuselage.PLANFORM_AREA]
        sys_press = inputs[Aircraft.Hydraulics.SYSTEM_PRESSURE]
        scaler = inputs[Aircraft.Hydraulics.MASS_SCALER]
        area = inputs[Aircraft.Wing.AREA]
        var_sweep = inputs[Aircraft.Wing.VAR_SWEEP_MASS_PENALTY]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        outputs[Aircraft.Hydraulics.MASS] = (
            0.57 * (planform + 0.27 * area)
            * (1 + 0.03 * num_wing_eng_fact + 0.05 * num_fuse_eng_fact)
            * (3000 / sys_press)**0.35 * (1 + 0.04 * var_sweep) * max_mach**0.33
            * scaler / GRAV_ENGLISH_LBM)

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_wing_eng = aviary_options.get_val(
            Aircraft.Propulsion.TOTAL_NUM_WING_ENGINES)
        num_fuse_eng = aviary_options.get_val(
            Aircraft.Propulsion.TOTAL_NUM_FUSELAGE_ENGINES)
        num_wing_eng_fact = distributed_engine_count_factor(num_wing_eng)
        num_fuse_eng_fact = distributed_engine_count_factor(num_fuse_eng)

        planform = inputs[Aircraft.Fuselage.PLANFORM_AREA]
        sys_press = inputs[Aircraft.Hydraulics.SYSTEM_PRESSURE]
        scaler = inputs[Aircraft.Hydraulics.MASS_SCALER]
        area = inputs[Aircraft.Wing.AREA]
        var_sweep = inputs[Aircraft.Wing.VAR_SWEEP_MASS_PENALTY]
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        term1 = (planform + 0.27 * area)
        term2 = (1.0 + 0.03 * num_wing_eng_fact + 0.05 * num_fuse_eng_fact)
        term3 = (3000.0 / sys_press)**0.35
        term4 = (1.0 + 0.04 * var_sweep)
        term5 = max_mach**0.33

        J[Aircraft.Hydraulics.MASS, Aircraft.Fuselage.PLANFORM_AREA] = (
            0.57 * term2 * term3 * term4 * term5 * scaler / GRAV_ENGLISH_LBM)

        J[Aircraft.Hydraulics.MASS, Aircraft.Wing.AREA] = (
            0.1539 * term2 * term3 * term4 * term5 * scaler / GRAV_ENGLISH_LBM)

        J[Aircraft.Hydraulics.MASS, Aircraft.Hydraulics.SYSTEM_PRESSURE] = (
            -3.2880267277063872637891549227565 * term1 * term2
            * sys_press**-1.35 * term4 * term5 * scaler / GRAV_ENGLISH_LBM)

        J[Aircraft.Hydraulics.MASS, Aircraft.Wing.VAR_SWEEP_MASS_PENALTY] = (
            0.0228 * term1 * term2 * term3 * term5 * scaler / GRAV_ENGLISH_LBM)

        J[Aircraft.Hydraulics.MASS, Aircraft.Hydraulics.MASS_SCALER] = (
            0.57 * term1 * term2 * term3 * term4 * term5 / GRAV_ENGLISH_LBM)


class AltHydraulicsGroupMass(om.ExplicitComponent):
    '''
    Calculates the mass of the hydraulics group using the alternate method.
    The methodology is based on the FLOPS weight equations, modified to
    output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Wing.AREA, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.WETTED_AREA, val=0.0)

        add_aviary_input(self, Aircraft.HorizontalTail.THICKNESS_TO_CHORD, val=0.0)

        add_aviary_input(self, Aircraft.VerticalTail.AREA, val=0.0)

        add_aviary_input(self, Aircraft.Hydraulics.MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.Hydraulics.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        area = inputs[Aircraft.Wing.AREA]
        horiz_wetted_area = inputs[Aircraft.HorizontalTail.WETTED_AREA]
        horiz_thick_chord = inputs[Aircraft.HorizontalTail.THICKNESS_TO_CHORD]
        vert_area = inputs[Aircraft.VerticalTail.AREA]
        scaler = inputs[Aircraft.Hydraulics.MASS_SCALER]

        outputs[Aircraft.Hydraulics.MASS] = (
            0.6053 * (area + 1.44 * (horiz_wetted_area /
                      (2.0 + 0.387 * horiz_thick_chord) + vert_area)) * scaler / GRAV_ENGLISH_LBM
        )

    def compute_partials(self, inputs, J):
        area = inputs[Aircraft.Wing.AREA]
        horiz_wetted_area = inputs[Aircraft.HorizontalTail.WETTED_AREA]
        horiz_thick_chord = inputs[Aircraft.HorizontalTail.THICKNESS_TO_CHORD]
        thick_chord_term = \
            2.0 + 0.387 * inputs[Aircraft.HorizontalTail.THICKNESS_TO_CHORD]
        vert_area = inputs[Aircraft.VerticalTail.AREA]
        scaler = inputs[Aircraft.Hydraulics.MASS_SCALER]

        J[Aircraft.Hydraulics.MASS, Aircraft.Wing.AREA] = (
            0.6053 * scaler / GRAV_ENGLISH_LBM
        )

        J[Aircraft.Hydraulics.MASS, Aircraft.HorizontalTail.WETTED_AREA] = (
            0.871632 / thick_chord_term * scaler / GRAV_ENGLISH_LBM
        )

        J[Aircraft.Hydraulics.MASS, Aircraft.HorizontalTail.THICKNESS_TO_CHORD] = (
            -0.337321584 * horiz_wetted_area /
            (thick_chord_term * thick_chord_term) * scaler / GRAV_ENGLISH_LBM
        )

        J[Aircraft.Hydraulics.MASS, Aircraft.VerticalTail.AREA] = (
            0.871632 * scaler / GRAV_ENGLISH_LBM
        )

        J[Aircraft.Hydraulics.MASS, Aircraft.Hydraulics.MASS_SCALER] = (
            0.6053 * (area + 1.44 * (horiz_wetted_area /
                      (2.0 + 0.387 * horiz_thick_chord) + vert_area)) / GRAV_ENGLISH_LBM
        )



import unittest
import os

import numpy as np
import openmdao
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials
from packaging import version

from aviary.mission.gasp_based.phases.landing_group import LandingSegment
from aviary.variable_info.options import get_option_defaults
from aviary.utils.test_utils.IO_test_util import check_prob_outputs
from aviary.variable_info.variables import Dynamic, Mission


class DLandTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model = LandingSegment(aviary_options=get_option_defaults())

    @unittest.skipIf(version.parse(openmdao.__version__) < version.parse("3.26"), "Skipping due to OpenMDAO version being too low (<3.26)")
    def test_dland(self):
        self.prob.setup(check=False, force_alloc_complex=True)

        self.prob.set_val(Mission.Landing.AIRPORT_ALTITUDE, 0, units="ft")
        self.prob.set_val(Mission.Landing.INITIAL_MACH, 0.1, units="unitless")
        self.prob.set_val("alpha", 0, units="deg")  # doesn't matter
        self.prob.set_val(Mission.Landing.MAXIMUM_SINK_RATE, 900, units="ft/min")
        self.prob.set_val(Mission.Landing.GLIDE_TO_STALL_RATIO, 1.3, units="unitless")
        self.prob.set_val(Mission.Landing.MAXIMUM_FLARE_LOAD_FACTOR,
                          1.15, units="unitless")
        self.prob.set_val(Mission.Landing.TOUCHDOWN_SINK_RATE, 5, units="ft/s")
        self.prob.set_val(Mission.Landing.BRAKING_DELAY, 1, units="s")
        self.prob.set_val("mass", 165279, units="lbm")
        self.prob.set_val(Dynamic.Mission.THROTTLE, 0.0, units='unitless')

        self.prob.run_model()

        testvals = {
            Mission.Landing.INITIAL_VELOCITY: 142.74 * 1.68781,
            "TAS_touchdown": 126.27 * 1.68781,
            "theta": np.deg2rad(3.57),
            "flare_alt": 20.8,
            "ground_roll_distance": 1798,
            Mission.Landing.GROUND_DISTANCE: 2980,
            "CL_max": 2.9533,
        }
        check_prob_outputs(self.prob, testvals, rtol=1e-2)

        partial_data = self.prob.check_partials(
            out_stream=None, method="cs", excludes=["*USatm*", "*params*", "*aero*"]
        )
        assert_check_partials(partial_data, atol=1e-6, rtol=1e-6)


if __name__ == "__main__":
    unittest.main()



## Depreciated Variables ##
# FLOPS
flops_deprecated_vars = [
    'AERIN.MODARO',
    'ENGDIN.IGENEN',
    'ENGDIN.BOOST',
    'ENGDIN.EXTFAC',
    'ENGDIN.IXTRAP',
    'ENGDIN.NPCODE',
    'ENGDIN.PCODE',
    'ENGDIN.NOX',
    'OPTION.IOPT',
    'OPTION.IPOLP',
    'OPTION.NOISE'
]

# GASP
gasp_deprecated_vars = []
